from pathlib import Path
import requests

try:
    import urllib.parse as urlparse  # Python 3
except ImportError:
    import urlparse  # Python 2

import argparse
import datetime
import logging
import os
import re
from io import BytesIO
import sys
import time

from zipfile import ZipFile

import yaml

requests.packages.urllib3.disable_warnings() 

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler(sys.stdout))
logger.setLevel(logging.INFO)

import pdb

# The regex to find the measurement id from the measurement page
# This should be read from the uploaded file, but would require an extra NetCDF module.
regex = "<h3>Measurement (?P<measurement_id>.{12,15}) <small>"  # {12, 15} to handle both old- and new-style measurement ids.


class SCC:
    """A simple class that will attempt to upload a file on the SCC server.

    The uploading is done by simulating a normal browser session. In the current
    version no check is performed, and no feedback is given if the upload
    was successful. If everything is setup correctly, it will work.
    """

    def __init__(self, auth, output_dir: Path | str | None, base_url: str):
        self.auth = auth
        self.base_url = base_url
        self.output_dir = self._set_output_dir(output_dir)
        self.session = requests.Session()
        self.session.auth = auth
        self.session.verify = False

        self.login_url: str = urlparse.urljoin(self.base_url, "accounts/login/")
        self.logout_url = urlparse.urljoin(self.base_url, "accounts/logout/")
        self.list_measurements_url = urlparse.urljoin(
            self.base_url, "data_processing/measurements/"
        )

        self.upload_url = urlparse.urljoin(
            self.base_url, "data_processing/measurements/quick/"
        )

        self.download_raw_pattern = urlparse.urljoin(
            self.base_url, "data_processing/measurements/{0}/download-raw/"
        )

        self.download_hirelpp_pattern = urlparse.urljoin(
            self.base_url, "data_processing/measurements/{0}/download-hirelpp/"
        )

        self.download_cloudmask_pattern = urlparse.urljoin(
            self.base_url, "data_processing/measurements/{0}/download-cloudmask/"
        )

        self.download_preprocessed_pattern = urlparse.urljoin(
            self.base_url, "data_processing/measurements/{0}/download-preprocessed/"
        )
        self.download_optical_pattern = urlparse.urljoin(
            self.base_url, "data_processing/measurements/{0}/download-optical/"
        )
        self.download_graph_pattern = urlparse.urljoin(
            self.base_url, "data_processing/measurements/{0}/download-plots/"
        )
        self.download_elic_pattern = urlparse.urljoin(
            self.base_url, "data_processing/measurements/{0}/download-elic/"
        )
        self.download_elquick_pattern = urlparse.urljoin(
            self.base_url, "data_processing/measurements/{0}/download-elquick/"
        )
        self.delete_measurement_pattern = urlparse.urljoin(
            self.base_url, "admin/database/measurements/{0}/delete/"
        )

        self.api_base_url = urlparse.urljoin(self.base_url, "api/v1/")
        self.api_measurement_pattern = urlparse.urljoin(
            self.api_base_url, "measurements/{0}/"
        )
        self.api_measurements_url = urlparse.urljoin(self.api_base_url, "measurements")
        self.api_sounding_search_pattern = urlparse.urljoin(
            self.api_base_url, "sounding_files/?filename={0}"
        )
        self.api_lidarratio_search_pattern = urlparse.urljoin(
            self.api_base_url, "lidarratio_files/?filename={0}"
        )
        self.api_overlap_search_pattern = urlparse.urljoin(
            self.api_base_url, "overlap_files/?filename={0}"
        )
        
    def _set_output_dir(self, output_dir: Path | str | None):
        if output_dir is None:
            _output_dir = Path.cwd()
        elif isinstance(output_dir, str):
            _output_dir = Path(output_dir)
        elif isinstance(output_dir, Path):
            _output_dir = output_dir
        return _output_dir

    def login(self, credentials: list[str]):
        """Login to SCC."""
        logger.debug("Attempting to login to SCC, username %s." % credentials[0])
        login_credentials = {"username": credentials[0], "password": credentials[1]}

        logger.debug("Accessing login page at %s." % self.login_url)

        # Get upload form
        login_page = self.session.get(self.login_url)

        if not login_page.ok:
            raise self.PageNotAccessibleError(
                "Could not access login pages. Status code %s" % login_page.status_code
            )

        logger.debug("Submitting credentials.")
        # Submit the login data
        login_submit = self.session.post(
            self.login_url,
            data=login_credentials,
            headers={
                "X-CSRFToken": login_page.cookies["csrftoken"],
                "referer": self.login_url,
            },
        )
        return login_submit

    def logout(self):
        """Logout from SCC"""
        return self.session.get(self.logout_url, stream=True)

    def upload_file(
        self, filename, system_id, rs_filename=None, ov_filename=None, lr_filename=None
    ):
        """Upload a filename for processing with a specific system. If the
        upload is successful, it returns the measurement id."""
        # Get submit page
        upload_page = self.session.get(self.upload_url)

        # Submit the data
        upload_data = {"system": system_id}
        files = {"data": open(filename, "rb")}
        if rs_filename is not None:
            ancillary_file, _ = self.get_ancillary(rs_filename, "sounding")
            if isinstance(ancillary_file, AncillaryFile):
                if ancillary_file.already_on_scc:
                    logger.warning(
                        "Sounding file {0.filename} already on the SCC with id {0.id}. Ignoring it.".format(
                            ancillary_file
                        )
                    )
                else:
                    logger.debug("Adding sounding file %s" % rs_filename)
                    files["sounding_file"] = open(rs_filename, "rb")

        if ov_filename is not None:
            ancillary_file, _ = self.get_ancillary(ov_filename, "overlap")
            if isinstance(ancillary_file, AncillaryFile):
                if ancillary_file.already_on_scc:
                    logger.warning(
                        "Overlap file {0.filename} already on the SCC with id {0.id}. Ignoring it.".format(
                            ancillary_file
                        )
                    )
                else:
                    logger.debug("Adding overlap file %s" % ov_filename)
                    files["overlap_file"] = open(ov_filename, "rb")

        if lr_filename is not None:
            ancillary_file, _ = self.get_ancillary(lr_filename, "lidarratio")
            if isinstance(ancillary_file, AncillaryFile):
                if ancillary_file.already_on_scc:
                    logger.warning(
                        "Lidar ratio file {0.filename} already on the SCC with id {0.id}. Ignoring it.".format(
                            ancillary_file
                        )
                    )
                else:
                    logger.debug("Adding lidar ratio file %s" % lr_filename)
                    files["lidar_ratio_file"] = open(lr_filename, "rb")

        logger.info("Uploading of file(s) %s started." % filename)

        upload_submit = self.session.post(
            self.upload_url,
            data=upload_data,
            files=files,
            headers={
                "X-CSRFToken": upload_page.cookies["csrftoken"],
                "referer": self.upload_url,
            },
        )

        if upload_submit.status_code != 200:
            logger.warning(
                "Connection error. Status code: %s" % upload_submit.status_code
            )
            return False

        # Check if there was a redirect to a new page.
        if upload_submit.url == self.upload_url:
            measurement_id = False
            logger.error(
                "Uploaded file(s) rejected! Try to upload manually to see the error."
            )
        else:
            measurement_id = re.findall(regex, upload_submit.text)[0]
            logger.info(
                "Successfully uploaded measurement with id %s." % measurement_id
            )

        return measurement_id

    def download_files(self, measurement_id, scc_type, download_url):
        """Downloads some files from the download_url to the specified
        SCC type. This method is used to download preprocessed file, optical
        files etc.
        """
        # TODO: Make downloading more robust (e.g. in case that files do not exist on server).
        # Get the file
        request = self.session.get(download_url, stream=True)

        if not request.ok:
            raise Exception(
                "Could not download files for measurement '%s'" % measurement_id
            )

        # Create the dir if it does not exist
        local_dir = self.output_dir
        local_dir.mkdir(parents=True, exist_ok=True)

        # Get the file name from the URL or provide a custom name
        scc_product = download_url.split("/")[-2].split("-")[-1]
        file_name = f"{scc_product}_{measurement_id}.zip"

        # Specify the local path for saving the file
        local_file_path = local_dir / file_name
        # Save the file by chunk, needed if the file is big
        with local_file_path.open("wb") as local_file:
            for chunk in request.iter_content(chunk_size=1024):
                if chunk:
                    local_file.write(chunk)
        logger.info(f"{scc_type} files downloaded.")

    def download_raw(self, measurement_id):
        """Download raw files for the measurement id."""
        # Construct the download url
        download_url = self.download_raw_pattern.format(measurement_id)
        self.download_files(measurement_id, "raw", download_url)

    def download_hirelpp(self, measurement_id):
        """Download HiRElPP files for the measurement id."""
        # Construct the download url
        download_url = self.download_hirelpp_pattern.format(measurement_id)
        self.download_files(measurement_id, "hirelpp", download_url)

    def download_cloudmask(self, measurement_id):
        """Download preprocessed files for the measurement id."""
        # Construct the download url
        download_url = self.download_cloudmask_pattern.format(measurement_id)
        self.download_files(measurement_id, "cloudmask", download_url)

    def download_preprocessed(self, measurement_id):
        """Download preprocessed files for the measurement id."""
        # Construct the download url
        download_url = self.download_preprocessed_pattern.format(measurement_id)
        self.download_files(measurement_id, "scc_preprocessed", download_url)

    def download_optical(self, measurement_id):
        """Download optical files for the measurement id."""
        # Construct the download url
        download_url = self.download_optical_pattern.format(measurement_id)
        self.download_files(measurement_id, "scc_optical", download_url)

    def download_graphs(self, measurement_id):
        """Download profile graphs for the measurement id."""
        # Construct the download url
        download_url = self.download_graph_pattern.format(measurement_id)
        self.download_files(measurement_id, "scc_plots", download_url)

    def download_elic(self, measurement_id):
        """Download profile graphs for the measurement id."""
        # Construct the download url
        download_url = self.download_elic_pattern.format(measurement_id)
        self.download_files(measurement_id, "elic", download_url)

    def download_elquick(self, measurement_id):
        """Download elquick for the measurement id."""
        # Construct the download url
        download_url = self.download_elquick_pattern.format(measurement_id)
        # self.download_files(measurement_id, 'elquick', download_url)
        logger.warning("Elquick download not available yet")

    def rerun_processing(self, measurement_id, monitor=True):
        measurement, status = self.get_measurement(measurement_id)

        if measurement:
            request = self.session.get(measurement.rerun_processing_url, stream=True)

            if request.status_code != 200:
                logger.error(
                    "Could not rerun processing for %s. Status code: %s"
                    % (measurement_id, request.status_code)
                )
                return

            if monitor:
                self.monitor_processing(measurement_id)

    def rerun_all(self, measurement_id, monitor=True):
        logger.debug("Started rerun_all procedure.")

        logger.debug("Getting measurement %s" % measurement_id)
        measurement, status = self.get_measurement(measurement_id)

        if measurement:
            logger.debug(
                "Attempting to rerun all processing through %s."
                % measurement.rerun_all_url
            )

            request = self.session.get(measurement.rerun_all_url, stream=True)

            if request.status_code != 200:
                logger.error(
                    "Could not rerun pre processing for %s. Status code: %s"
                    % (measurement_id, request.status_code)
                )
                return

            if monitor:
                self.monitor_processing(measurement_id)

    def process(
        self,
        filename,
        system_id,
        monitor,
        rs_filename=None,
        lr_filename=None,
        ov_filename=None,
    ):
        """Upload a file for processing and wait for the processing to finish.
        If the processing is successful, it will download all produced files.
        """
        logger.info("--- Processing started on %s. ---" % datetime.datetime.now())
        # Upload file
        logger.info("--- Uploading file")
        measurement_id = self.upload_file(
            filename,
            system_id,
            rs_filename=rs_filename,
            lr_filename=lr_filename,
            ov_filename=ov_filename,
        )

        if measurement_id and monitor:
            logger.info("--- Monitoring processing")
            return self.monitor_processing(measurement_id)

        return None

    def monitor_processing(
        self,
        measurement_id: str,
        download_raw: bool = False,
        download_hirelpp: bool = False,
        download_cloudmask: bool = False,
        download_elpp: bool =True,
        download_elda: bool =True,
        download_elic: bool = False,
        download_elquick: bool = False,
        download_eldec: bool = False,
        upload_max: int = 10,
        time_sleep_upload: int = 10,
        queue_max: int = 10,
        time_sleep_queue: int = 10,
        run_max: int = 20,
        time_sleep_run: int = 10,
    ):
        """Monitor the processing progress of a measurement id. If the download flags are set to True, the files will be downloaded when the processing is finished.

        Args:
            measurement_id (str): Measurement id.
            download_raw (bool, optional): raw files will be downloared. Defaults to False.
            download_hirelpp (bool, optional): hirelpp files will be downloared. Defaults to False.
            download_cloudmask (bool, optional): cloudmask files will be downloared. Defaults to False.
            download_elpp (bool, optional): elpp files will be downloared. Defaults to True.
            download_elda (bool, optional): elda files will be downloared. Defaults to True.
            download_elic (bool, optional): elic files will be downloared. Defaults to False.
            download_elquick (bool, optional): elquick files will be downloared. Defaults to False.
            download_eldec (bool, optional): elded files will be downloared. Defaults to False.
            upload_max (int, optional): Maximum number of times of `time_sleep_upload` will wait file to be uploaded to the SCC. Default 10.
            time_sleep_upload (int, optional): Time in seconds to wait file to be uploaded to the SCC. Default 10 seconds. 
            queue_max (int, optional): Maximum number of times of `time_sleep_queue` will wait file to leave queue. Default 10.
            time_sleep_queue (int, optional): Time in seconds to wait file to leave queue to the SCC. Default 10 seconds. 
            run_max (int, optional): Maximum number of times of `time_sleep_run` will wait file to finish processing. Default 10.
            time_sleep_run (int, optional): Time in seconds to wait file to finish processing. Default 10 seconds. 

        Raises:
            TimeoutError: File not found in the SCC DB.
            Warning: Measurement is delayed.
        """
        upload_count, queue_count, run_count = 0, 0, 0

        # try to wait for measurement to appear in API
        measurement = None
        logger.info("Looking for measurement %s in SCC", measurement_id)
        while upload_count < upload_max:
            time.sleep(time_sleep_upload)
            measurement, status = self.get_measurement(measurement_id)
            if status == 200:
                logger.error(f"Measurement found: {measurement_id}")
                break
            else:
                if upload_count < upload_max:
                    logger.error("Measurement not found yet. waiting %ds", time_sleep_upload)
                    upload_count += 1
                else:
                    raise TimeoutError(f"Measurement {measurement_id} doesn't seem to exist.")

        if measurement is None:
            raise FileNotFoundError(f"Measurement {measurement_id} doesn't seem to exist.")
            
        # Start download of hirelpp files
        products = [
            "upload",
            "hirelpp",
            "cloudmask",
            "elpp",
            "elda",
            "elquick",
            "elic",
            "eldec",
        ]
        measurement, status = self.get_measurement(measurement_id)

        if measurement is None:
            raise FileNotFoundError(f"Measurement {measurement_id} doesn't seem to exist.")

        status = [
            measurement.upload, #raw 
            measurement.hirelpp,
            measurement.cloudmask,
            measurement.elpp,
            measurement.elda,
            measurement.elquick,
            measurement.elic,
            measurement.eldec,
        ]
        status_dict = dict(zip(products, status))

        required_downloads = [
            download_raw,
            download_hirelpp,
            download_cloudmask,
            download_elpp,
            download_elda,
            download_elquick,
            download_elic,
            download_eldec,
        ]
        required_products = [
            product
            for product, required in zip(products, required_downloads)
            if required
        ]
        
        if self.is_being_processed(measurement_id):
            logger.info(f"Measurement {measurement_id} is being processed")
        elif self.is_queued(measurement_id):
            logger.info(f"Measurement {measurement_id} is queued")
        elif self.is_delayed(measurement_id):
            raise Warning(f"Measurement {measurement_id} is delayed")


        # Wait for processing to finish
        if self.is_queued(measurement_id):
            while (queue_count <= queue_max):
                logger.info(f"#{queue_count:02d}/{queue_max}: Process is queued. Please wait. ")
                time.sleep(time_sleep_queue)
                measurement, status = self.get_measurement(measurement_id)
                if measurement is None:
                    raise FileNotFoundError(f"Measurement {measurement_id} doesn't seem to exist.")
                                    
                if not measurement.is_queued:
                    logger.info("Process is running.")
                    break
                else:
                    queue_count += 1
            
            if queue_count > queue_max:
                raise TimeoutError('Process was queued too long.')

        if self.is_being_processed(measurement_id):
            while (run_count <= run_max):
                info_str = f"RAW: {measurement.upload}, HIRELPP: {measurement.hirelpp}, CLOUDMASK: {measurement.cloudmask}, ELPP: {measurement.elpp}, ELDA: {measurement.elda}, ELIC: {measurement.elic}, ELQUICK: {measurement.elquick}"
                logger.info(f"#{run_count:02d}/{run_max}: Please wait. " + info_str)
                
                time.sleep(time_sleep_run)
                
                measurement, status = self.get_measurement(measurement_id)
                
                if measurement is None:
                    raise FileNotFoundError(f"Measurement {measurement_id} doesn't seem to exist.")            
                
                if not self.is_being_processed(measurement_id):
                    logger.info("--- SCC processing finished. ---")
                    break

                if all([getattr(measurement, product)==127 for product in required_products]):
                    logger.info(f"All SCC products to be downloaded have been already processed by the SCC.")
                    break
                elif getattr(measurement, "elpp") == -127:
                    logger.warning(f"ELPP error code {measurement.elpp_exit_code['exit_code']}: {measurement.elpp_exit_code['description']}")
                    break
                run_count += 1
            
            if run_count == run_max:
                logger.warning(f"Timeout reached. Not all products were processed.")
                return measurement

        for product in required_products:            
            if status_dict[product] == 127:
                if product == "upload":
                    self.download_files(
                        measurement_id,
                        product,
                        self.download_raw_pattern.format(measurement_id),
                    )
                if product == "hirelpp":
                    self.download_files(
                        measurement_id,
                        product,
                        self.download_hirelpp_pattern.format(measurement_id),
                    )
                elif product == "cloudmask":
                    self.download_files(
                        measurement_id,
                        product,
                        self.download_cloudmask_pattern.format(measurement_id),
                    )
                elif product == "elpp":
                    self.download_files(
                        measurement_id,
                        product,
                        self.download_preprocessed_pattern.format(measurement_id),
                    )
                elif product == "elda":
                    self.download_files(
                        measurement_id,
                        product,
                        self.download_optical_pattern.format(measurement_id),
                    )
                elif product == "elquick":
                    self.download_files(
                        measurement_id,
                        product,
                        self.download_elquick_pattern.format(measurement_id),
                    )
                elif product == "elic":
                    self.download_files(
                        measurement_id,
                        product,
                        self.download_elic_pattern.format(measurement_id),
                    )
                elif product == "eldec":
                    logger.info(f"ELDEC downloading not implemented yet.")
                    # self.download_files(measurement_id, product, self.download_eldec_pattern.format(measurement_id))
            elif status_dict[product] == -127:
                logger.warning(f"{product} not avaiable in the SCC.")
                if product == "hirelpp":
                    logger.warning(
                        f"{product} error code {measurement.hirelpp_exit_code['exit_code']}: {measurement.hirelpp_exit_code['description']}"
                    )
                elif product == "cloudmask":
                    logger.warning(
                        f"{product} error code {measurement.cloudmask_exit_code['exit_code']}: {measurement.cloudmask_exit_code['description']}"
                    )
                elif product == "elpp":
                    logger.warning(
                        f"{product} error code {measurement.elpp_exit_code['exit_code']}: {measurement.elpp_exit_code['description']}"
                    )
                elif product == "elda":
                    logger.warning(
                        f"{product} error code {measurement.elda_exit_code['exit_code']}: {measurement.elda_exit_code['description']}"
                    )
                elif product == "elquick":
                    logger.warning(
                        f"{product} error code {measurement.elquick_exit_code['exit_code']}: {measurement.elquick_exit_code['description']}"
                    )
                elif product == "elic":
                    logger.warning(
                        f"{product} error code {measurement.elic_exit_code['exit_code']}: {measurement.elic_exit_code['description']}"
                    )
                elif product == "eldec":
                    logger.warning(
                        f"{product} error code {measurement.eldec_exit_code['exit_code']}: {measurement.eldec_exit_code['description']}"
                    )
                else:
                    logger.warning(f"Unknown error code for {product}: {status_dict[product]}")

    def get_measurement(self, measurement_id):
        measurement_url = self.api_measurement_pattern.format(measurement_id)
        logger.debug("Measurement API URL: %s" % measurement_url)

        response = self.session.get(measurement_url)
        if not response.ok:
            logger.error("Could not access API. Status code %s." % response.status_code)
            return None, response.status_code

        response_dict = response.json()
    
        if response_dict:
            measurement = Measurement(self.base_url, response_dict)
            return measurement, response.status_code
        else:
            logger.error("No measurement with id %s found on the SCC." % measurement_id)
            return None, response.status_code

    def is_being_processed(self, measurement_id):
        measurement_url = self.api_measurement_pattern.format(measurement_id)
        logger.debug("Measurement API URL: %s" % measurement_url)

        response = self.session.get(measurement_url)
        if not response.ok:
            logger.error("Could not access API. Status code %s." % response.status_code)
            return None, response.status_code
        else:
            response_dict = response.json()
            return response_dict["is_being_processed"]

    def is_delayed(self, measurement_id):
        measurement_url = self.api_measurement_pattern.format(measurement_id)
        logger.debug("Measurement API URL: %s" % measurement_url)

        response = self.session.get(measurement_url)
        if not response.ok:
            logger.error("Could not access API. Status code %s." % response.status_code)
            return None, response.status_code
        else:
            response_dict = response.json()
            return response_dict["is_delayed"]

    def is_queued(self, measurement_id):
        measurement_url = self.api_measurement_pattern.format(measurement_id)
        logger.debug("Measurement API URL: %s" % measurement_url)

        response = self.session.get(measurement_url)
        if not response.ok:
            logger.error("Could not access API. Status code %s." % response.status_code)
            return None, response.status_code
        else:
            response_dict = response.json()
            return response_dict["is_queued"]

    def delete_measurement(self, measurement_id: str):
        """Deletes a measurement with the provided measurement id. The user
        should have the appropriate permissions.

        The procedures is performed directly through the web interface and
        NOT through the API.
        """
        # Get the measurement object
        measurement, _ = self.get_measurement(measurement_id)

        # Check that it exists
        if measurement is None:
            logger.warning("Nothing to delete.")
            return None

        # Go the the page confirming the deletion

        delete_url = self.delete_measurement_pattern.format(measurement_id)

        confirm_page = self.session.get(delete_url)
        # Check that the page opened properly
        if confirm_page.status_code != 200:
            logger.warning(
                "Could not open delete page. Status: {0}".format(
                    confirm_page.status_code
                )
            )
            return None

        # Delete the measurement
        delete_page = self.session.post(
            delete_url,
            data={"post": "yes"},
            headers={
                "X-CSRFToken": confirm_page.cookies["csrftoken"],
                "referer": delete_url,
            },
        )

        if not delete_page.ok:
            logger.warning(
                "Something went wrong. Delete page status: {0}".format(
                    delete_page.status_code
                )
            )
            return None

        logger.info("Deleted measurement {0}".format(measurement_id))
        return True

    def available_measurements(self):
        """Get a list of available measurement on the SCC."""

        response = self.session.get(self.api_measurements_url)
        response_dict = response.json()

        if response_dict:
            measurement_list = response_dict["objects"]
            measurements = [
                Measurement(self.base_url, measurement_dict)
                for measurement_dict in measurement_list
            ]
            logger.info("Found %s measurements on the SCC." % len(measurements))
        else:
            logger.warning(
                "No response received from the SCC when asked for available measurements."
            )
            measurements = None

        return measurements

    def list_measurements(
        self,
        station=None,
        system=None,
        start=None,
        stop=None,
        upload_status=None,
        processing_status=None,
        optical_processing=None,
    ):
        # TODO: Change this to work through the API

        # Need to set to empty string if not specified, we won't get any results
        params = {
            "station": station if station is not None else "",
            "system": system if system is not None else "",
            "stop": stop if stop is not None else "",
            "start": start if start is not None else "",
            "upload_status": upload_status if upload_status is not None else "",
            "preprocessing_status": processing_status
            if processing_status is not None
            else "",
            "optical_processing_status": optical_processing
            if optical_processing is not None
            else "",
        }

        response_txt = self.session.get(self.list_measurements_url, params=params).text
        tbl_rgx = re.compile(r'<table id="measurements">(.*?)</table>', re.DOTALL)
        entry_rgx = re.compile(r"<tr>(.*?)</tr>", re.DOTALL)
        measurement_rgx = re.compile(
            r'.*?<td><a[^>]*>(\w+)</a>.*?<td>.*?<td>([\w-]+ [\w:]+)</td>.*<td data-order="([-]?\d+),([-]?\d+),([-]?\d+)".*',
            re.DOTALL,
        )
        matches = tbl_rgx.findall(response_txt)
        if len(matches) != 1:
            return []

        ret = []
        for entry in entry_rgx.finditer(matches[0]):
            m = measurement_rgx.match(entry.string[entry.start(0) : entry.end(0)])
            if m:
                name, date, upload, preproc, optical = m.groups()
                ret.append(
                    Measurement(
                        self.base_url,
                        {
                            "id": name,
                            "upload": int(upload),
                            "pre_processing": int(preproc),
                            "processing": int(optical),
                        },
                    )
                )

        return ret

    def measurement_id_for_date(self, t1, call_sign, base_number=0):
        """Give the first available measurement id on the SCC for the specific
        date.
        """
        date_str = t1.strftime("%Y%m%d")
        base_id = "%s%s" % (date_str, call_sign)
        search_url = urlparse.urljoin(
            self.api_base_url, "measurements/?id__startswith=%s" % base_id
        )

        response = self.session.get(search_url)

        response_dict = response.json()

        measurement_id = None

        if response_dict:
            measurement_list = response_dict["objects"]

            if len(measurement_list) == 100:
                raise ValueError("No available measurement id found.")

            existing_ids = [
                measurement_dict["id"] for measurement_dict in measurement_list
            ]

            measurement_number = base_number
            measurement_id = "%s%02i" % (base_id, measurement_number)

            while measurement_id in existing_ids:
                measurement_number = measurement_number + 1
                measurement_id = "%s%02i" % (base_id, measurement_number)

        return measurement_id

    def get_ancillary(self, file_path, file_type):
        """
        Try to get the ancillary file data from the SCC API.

        The result will always be an API object. If the file does not exist, the .exists property is set to False.

        Parameters
        ----------
        file_path : str
           Path  of the uploaded file.
        file_type : str
           Type of ancillary file. One of 'sounding', 'overlap', 'lidarratio'.

        Returns
        : AncillaryFile
           The api object.
        """
        assert file_type in ["sounding", "overlap", "lidarratio"]

        filename = os.path.basename(file_path)

        if file_type == "sounding":
            file_url = self.api_sounding_search_pattern.format(filename)
        elif file_type == "overlap":
            file_url = self.api_overlap_search_pattern.format(filename)
        else:
            file_url = self.api_lidarratio_search_pattern.format(filename)

        response = self.session.get(file_url)

        if not response.ok:
            logger.error("Could not access API. Status code %s." % response.status_code)
            return None, response.status_code

        response_dict = response.json()
        object_list = response_dict["objects"]

        logger.debug("Ancillary file JSON: {0}".format(object_list))

        if object_list:
            ancillary_file = AncillaryFile(
                self.api_base_url, response_dict #TODO:Check if response_dict should be object_list
            )  # Assume only one file is returned
        else:
            ancillary_file = AncillaryFile(
                self.api_base_url, None
            )  # Create an empty object

        return ancillary_file, response.status_code

    class PageNotAccessibleError(RuntimeError):
        pass


class ApiObject(object):
    """A generic class object."""

    def __init__(self, base_url, dict_response):
        self.base_url = base_url

        if dict_response:
            # Add the dictionary key value pairs as object properties
            for key, value in dict_response.items():
                # logger.debug('Setting key {0} to value {1}'.format(key, value))
                try:
                    setattr(self, key, value)
                except:
                    logger.warning(
                        "Could not set attribute {0} to value {1}".format(key, value)
                    )
            self.exists = True
        else:
            self.exists = False


class Measurement(ApiObject):
    """This class represents the measurement object as returned in the SCC API."""
    def __init__(self, base_url, dict_response):
        super(Measurement, self).__init__(base_url, dict_response)
        self.id = self.id
        self.upload = self.upload 
        self.hirelpp = self.hirelpp
        self.cloudmask = self.cloudmask
        self.elpp = self.elpp
        self.elda = self.elda
        self.elquick = self.elquick
        self.elic = self.elic
        self.eldec = self.eldec
        self.is_being_processed: bool = self.is_being_processed
        self.is_delayed: bool = self.is_delayed
        self.is_queued: bool = self.is_queued
        self.pre_processing: bool = False
        self.processing: bool = False
        self.elpp_exit_code: dict = self.elpp_exit_code
        self.elda_exit_code: dict = self.elda_exit_code
        self.elquick_exit_code: dict = self.elquick_exit_code
        self.elic_exit_code: dict = self.elic_exit_code
        self.eldec_exit_code: dict = self.eldec_exit_code
        self.hirelpp_exit_code: dict = self.hirelpp_exit_code
        self.cloudmask_exit_code: dict = self.cloudmask_exit_code  


    @property
    def rerun_processing_url(self):
        url_pattern = urlparse.urljoin(
            self.base_url, "data_processing/measurements/{0}/rerun-elda/"
        )
        return url_pattern.format(self.id) #type: ignore

    @property
    def rerun_all_url(self):
        ulr_pattern = urlparse.urljoin(
            self.base_url, "data_processing/measurements/{0}/rerun-all/"
        )
        return ulr_pattern.format(self.id) #type: ignore

    def __str__(self):
        return "%s: %s, %s, %s" % (
            self.id, #type: ignore
            self.upload, #type: ignore
            self.pre_processing,
            self.processing,
        )


class AncillaryFile(ApiObject):
    """This class represents the ancilalry file object as returned in the SCC API."""
    def __init__(self, base_url, dict_response):
        super(AncillaryFile, self).__init__(base_url, dict_response)

    @property
    def already_on_scc(self):        
        return self.exists

    def __str__(self):
        return "%s: %s, %s" % (self.id, self.filename, self.status) #type: ignore


def process_file(
    filename,
    system_id,
    settings,
    monitor=True,
    rs_filename=None,
    lr_filename=None,
    ov_filename=None,
):
    """Shortcut function to process a file to the SCC."""
    logger.info("Processing file %s, using system %s" % (filename, system_id))

    scc = SCC(
        settings["basic_credentials"], settings["output_dir"], settings["base_url"]
    )
    scc.login(settings["website_credentials"])
    measurement = scc.process(
        filename,
        system_id,
        monitor=monitor,
        rs_filename=rs_filename,
        lr_filename=lr_filename,
        ov_filename=ov_filename,
    )
    scc.logout()
    return measurement


def delete_measurements(measurement_ids, settings):
    """Shortcut function to delete measurements from the SCC."""
    scc = SCC(
        settings["basic_credentials"], settings["output_dir"], settings["base_url"]
    )
    scc.login(settings["website_credentials"])
    for m_id in measurement_ids:
        logger.info("Deleting %s" % m_id)
        scc.delete_measurement(m_id)
    scc.logout()


def rerun_all(measurement_ids, monitor, settings):
    """Shortcut function to rerun measurements from the SCC."""

    scc = SCC(
        settings["basic_credentials"], settings["output_dir"], settings["base_url"]
    )
    scc.login(settings["website_credentials"])
    for m_id in measurement_ids:
        logger.info("Rerunning all products for %s" % m_id)
        scc.rerun_all(m_id, monitor)
    scc.logout()


def rerun_processing(measurement_ids, monitor, settings):
    """Shortcut function to delete a measurement from the SCC."""

    scc = SCC(
        settings["basic_credentials"], settings["output_dir"], settings["base_url"]
    )
    scc.login(settings["website_credentials"])
    for m_id in measurement_ids:
        logger.info("Rerunning (optical) processing for %s" % m_id)
        scc.rerun_processing(m_id, monitor)
    scc.logout()


def list_measurements(
    settings,
    station=None,
    system=None,
    start=None,
    stop=None,
    upload_status=None,
    preprocessing_status=None,
    optical_processing=None,
):
    """List all available measurements"""
    scc = SCC(
        settings["basic_credentials"], settings["output_dir"], settings["base_url"]
    )
    scc.login(settings["website_credentials"])
    ret = scc.list_measurements(
        station=station,
        system=system,
        start=start,
        stop=stop,
        upload_status=upload_status,
        processing_status=preprocessing_status,
        optical_processing=optical_processing,
    )
    for entry in ret:
        print("%s" % entry.id)
    scc.logout()


def download_measurements(
    measurement_ids, download_preproc, download_optical, download_graph, settings
):
    """Download all measurements for the specified IDs"""
    scc = SCC(
        settings["basic_credentials"], settings["output_dir"], settings["base_url"]
    )
    scc.login(settings["website_credentials"])
    for m_id in measurement_ids:
        if download_preproc:
            logger.info("Downloading preprocessed files for '%s'" % m_id)
            scc.download_preprocessed(m_id)
            logger.info("Complete")
        if download_optical:
            logger.info("Downloading optical files for '%s'" % m_id)
            scc.download_optical(m_id)
            logger.info("Complete")
        if download_graph:
            logger.info("Downloading profile graph files for '%s'" % m_id)
            scc.download_graphs(m_id)
            logger.info("Complete")


def settings_from_path(config_file_path):
    """Read the configuration file.

    The file should be in YAML syntax."""

    if not os.path.isfile(config_file_path):
        raise argparse.ArgumentTypeError(
            "Wrong path for configuration file (%s)" % config_file_path
        )

    with open(config_file_path) as yaml_file:
        try:
            settings = yaml.safe_load(yaml_file)
            logger.debug("Read settings file(%s)" % config_file_path)
        except Exception:
            raise argparse.ArgumentTypeError(
                "Could not parse YAML file (%s)" % config_file_path
            )

    # YAML limitation: does not read tuples
    settings["basic_credentials"] = tuple(settings["basic_credentials"])
    settings["website_credentials"] = tuple(settings["website_credentials"])
    return settings


# Setup for command specific parsers
def setup_delete(parser):
    def delete_from_args(parsed):
        delete_measurements(parsed.IDs, parsed.config)

    parser.add_argument("IDs", nargs="+", help="measurement IDs to delete.")
    parser.set_defaults(execute=delete_from_args)


def setup_rerun_all(parser):
    def rerun_all_from_args(parsed):
        rerun_all(parsed.IDs, parsed.process, parsed.config)

    parser.add_argument("IDs", nargs="+", help="Measurement IDs to rerun.")
    parser.add_argument(
        "-p",
        "--process",
        help="Wait for the results of the processing.",
        action="store_true",
    )
    parser.set_defaults(execute=rerun_all_from_args)


def setup_rerun_processing(parser):
    def rerun_processing_from_args(parsed):
        rerun_processing(parsed.IDs, parsed.process, parsed.config)

    parser.add_argument(
        "IDs", nargs="+", help="Measurement IDs to rerun the processing on."
    )
    parser.add_argument(
        "-p",
        "--process",
        help="Wait for the results of the processing.",
        action="store_true",
    )
    parser.set_defaults(execute=rerun_processing_from_args)


def setup_process_file(parser):
    """Upload and monitor processing progress."""

    def process_file_from_args(parsed):
        process_file(
            parsed.filename,
            parsed.system,
            parsed.config,
            monitor=True,
            rs_filename=parsed.radiosounding,
            ov_filename=parsed.overlap,
            lr_filename=parsed.lidarratio,
        )

    parser.add_argument("filename", help="Measurement file name or path.")
    parser.add_argument("system", help="Processing system id.")
    parser.add_argument(
        "--radiosounding", default=None, help="Radiosounding file name or path"
    )
    parser.add_argument("--overlap", default=None, help="Overlap file name or path")
    parser.add_argument(
        "--lidarratio", default=None, help="Lidar ratio file name or path"
    )

    parser.set_defaults(execute=process_file_from_args)


def setup_upload_file(parser):
    """Upload but do not monitor processing progress."""

    def upload_file_from_args(parsed):
        process_file(
            parsed.filename,
            parsed.system,
            parsed.config,
            monitor=False,
            rs_filename=parsed.radiosounding,
            ov_filename=parsed.overlap,
            lr_filename=parsed.lidarratio,
        )

    parser.add_argument("filename", help="Measurement file name or path.")
    parser.add_argument("system", help="Processing system id.")
    parser.add_argument(
        "--radiosounding", default=None, help="Radiosounding file name or path"
    )
    parser.add_argument("--overlap", default=None, help="Overlap file name or path")
    parser.add_argument(
        "--lidarratio", default=None, help="Lidar ratio file name or path"
    )

    parser.set_defaults(execute=upload_file_from_args)


def setup_list_measurements(parser):
    def list_measurements_from_args(parsed):
        list_measurements(
            parsed.config,
            station=parsed.station,
            system=parsed.system,
            start=parsed.start,
            stop=parsed.stop,
            upload_status=parsed.upload_status,
            preprocessing_status=parsed.preprocessing_status,
            optical_processing=parsed.optical_processing_status,
        )

    def status(arg):
        if -127 <= int(arg) <= 127:
            return arg
        else:
            raise argparse.ArgumentTypeError("Status must be between -127 and 127")

    def date(arg):
        if re.match(r"\d{4}-\d{2}-\d{2}", arg):
            return arg
        else:
            raise argparse.ArgumentTypeError("Date must be in format 'YYYY-MM-DD'")

    parser.add_argument("--station", help="Filter for only the selected station")
    parser.add_argument("--system", help="Filter for only the selected station")
    parser.add_argument(
        "--start", help="Filter for only the selected station", type=date
    )
    parser.add_argument(
        "--stop", help="Filter for only the selected station", type=date
    )
    parser.add_argument(
        "--upload-status", help="Filter for only the selected station", type=status
    )
    parser.add_argument(
        "--preprocessing-status",
        help="Filter for only the selected station",
        type=status,
    )
    parser.add_argument(
        "--optical-processing-status",
        help="Filter for only the selected station",
        type=status,
    )
    parser.set_defaults(execute=list_measurements_from_args)


def setup_download_measurements(parser):
    def download_measurements_from_args(parsed):
        preproc = parsed.download_preprocessed
        optical = parsed.download_optical
        graphs = parsed.download_profile_graphs
        if not preproc and not graphs:
            optical = True
        download_measurements(parsed.IDs, preproc, optical, graphs, parsed.config)

    parser.add_argument(
        "IDs", help="Measurement IDs that should be downloaded.", nargs="+"
    )
    parser.add_argument(
        "--download-preprocessed",
        action="store_true",
        help="Download preprocessed files.",
    )
    parser.add_argument(
        "--download-optical",
        action="store_true",
        help="Download optical files (default if no other download is used).",
    )
    parser.add_argument(
        "--download-profile-graphs",
        action="store_true",
        help="Download profile graph files.",
    )
    parser.set_defaults(execute=download_measurements_from_args)


def main():
    # Define the command line arguments.
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()

    delete_parser = subparsers.add_parser("delete", help="Deletes a measurement.")
    rerun_all_parser = subparsers.add_parser(
        "rerun-all", help="Reprocess a measurement on the SCC."
    )
    rerun_processing_parser = subparsers.add_parser(
        "rerun-processing", help="Rerun processing routines for a measurement."
    )
    process_file_parser = subparsers.add_parser(
        "process-file", help="Upload a file and download processing results."
    )
    upload_file_parser = subparsers.add_parser("upload-file", help="Upload a file.")
    list_parser = subparsers.add_parser(
        "list", help="List measurements registered on the SCC."
    )
    download_parser = subparsers.add_parser(
        "download", help="Download selected measurements."
    )

    setup_delete(delete_parser)
    setup_rerun_all(rerun_all_parser)
    setup_rerun_processing(rerun_processing_parser)
    setup_process_file(process_file_parser)
    setup_upload_file(upload_file_parser)
    setup_list_measurements(list_parser)
    setup_download_measurements(download_parser)

    # Verbosity settings from http://stackoverflow.com/a/20663028
    parser.add_argument(
        "-d",
        "--debug",
        help="Print debugging information.",
        action="store_const",
        dest="loglevel",
        const=logging.DEBUG,
        default=logging.INFO,
    )
    parser.add_argument(
        "-s",
        "--silent",
        help="Show only warning and error messages.",
        action="store_const",
        dest="loglevel",
        const=logging.WARNING,
    )

    # Setup default config location
    home = os.path.expanduser("~")
    default_config_location = os.path.abspath(os.path.join(home, ".scc_access.yaml"))
    parser.add_argument(
        "-c",
        "--config",
        help="Path to the config file.",
        type=settings_from_path,
        default=default_config_location,
    )

    args = parser.parse_args()

    # Get the logger with the appropriate level
    logging.basicConfig(format="%(levelname)s: %(message)s", level=args.loglevel)

    # Dispatch to appropriate function
    args.execute(args)


# When running through terminal
if __name__ == "__main__":
    main()



from pathlib import Path

from datetime import datetime
import xarray as xr
import numpy as np
import pandas as pd
from typing import Any

from gfatpy.lidar.utils.utils import LIDAR_INFO
from gfatpy.atmo.ecmwf import get_ecmwf_temperature_pressure
from gfatpy.atmo.atmo import generate_meteo_profiles
from gfatpy.lidar.utils.file_manager import channel2info


def rayleigh2earlinet(dataset: xr.Dataset, output_dir: Path | str | None) -> tuple[Path, Path]:
    """It converts a Rayleigh Fit dataset to EARLINET format.

    Args:
        dataset (xr.Dataset): Rayleigh Fit dataset.
        output_dir (Path | str | None): Output directory. Defaults to None.

    Raises:
        NotADirectoryError: Directory not found.

    Returns:
        tuple[Path, Path]: NetCDF and CSV filepaths. 
    """    
    # info from dataset
    date = datetime.strptime(
        dataset.attrs["datetime_ini"], dataset.attrs["datetime_format"]
    )
    date_str = datetime.strftime(date, "%d.%m.%Y, %HUTC")
    lidar_name: str = dataset.attrs["lidar_name"].lower()
    lidar_location: str = dataset.attrs["lidar_location"]
    lidar_id: str = dataset.attrs["lidar_id"]
    channel: str = dataset.attrs["channel"]
    detection_mode: str = dataset["detection_mode"].values.item()
    if detection_mode == "a":
        dark_subtracted_str: str = dataset.attrs["dark_subtracted"]
    else:
        dark_subtracted_str: str = ""
    duration = float(dataset.attrs["duration"])
    wavelength, _, polarization_, mode_ = channel2info(channel)
    radiosonde_location = dataset.attrs["radiosonde_location"]
    if dataset.attrs["radiosonde_wmo_id"] is not None:
        radiosonde_wmo_id = dataset.attrs["radiosonde_wmo_id"]
    else:
        radiosonde_wmo_id = -9999
        dataset.attrs["radiosonde_wmo_id"] = radiosonde_wmo_id

    radiosonde_date = datetime.strptime(
        dataset.attrs["radiosonde_datetime"], "%Y-%m-%dT%H-%M-%S"
    )

    pol_str = LIDAR_INFO["metadata"]["code_polarization_str2long_name"][polarization_]  # type: ignore
    mode_str = LIDAR_INFO["metadata"]["code_mode_str2long_name"][mode_]  # type: ignore
    z_min, z_max = dataset.attrs["rayleigh_height_limits"]

    # create output_dir
    if output_dir is None:
        output_dir = Path.cwd()
    elif isinstance(output_dir, str):
        output_dir = Path(output_dir)

    if not output_dir.exists() or not output_dir.is_dir():
        raise NotADirectoryError(f"{output_dir} not found.")

    output_dir.mkdir(parents=True, exist_ok=True)

    rf_nc_fn = output_dir / f"{lidar_id}RayleighFit{channel}.nc"
    dataset.to_netcdf(rf_nc_fn)

    # # Filename
    rf_fn = output_dir / f"{lidar_id}RayleighFit{channel}.csv"

    # Select Columns to write
    cols = ["BCS", "RCS"]
    if detection_mode == "a":  # (if analog)
        cols.append("DC")
    rf_df = dataset[cols].to_dataframe()
    rf_df.columns = [dataset[col].attrs["name"] for col in cols]

    # Write File Earlinet Format
    with open(rf_fn, "w") as f:
        f.write(f"station ID = {lidar_id} ({lidar_location})\n")
        f.write(f"system = {lidar_name}\n")
        f.write(
            f"signal = {wavelength}, {pol_str}, {mode_str}, {dark_subtracted_str}\n"
        )
        f.write(
            f"date of measurement, time, duration of measurement= {date_str}, {duration:.1f} s\n"
        )
        f.write(
            f"location, WMO radiosonde station ID, date of radiosonde = {radiosonde_location}, {radiosonde_wmo_id}, {radiosonde_date}\n"
        )
        f.write(
            f"lower and upper Rayleigh height limits = {np.round(z_min)}, {np.round(z_max)}\n"
        )
    f.close()

    # write in the same file the rest of information
    rf_df.index = rf_df.index.map(lambda x: "%.4f" % x)
    rf_df.to_csv(rf_fn, mode="a", header=True, na_rep="NaN", float_format="%.4e")
    
    #Close file
    del rf_df
    return rf_nc_fn, rf_fn


def get_meteo(
    date: datetime,
    range: np.ndarray[Any, np.dtype[np.float64]],
    meteorology_source: str,
    pressure_surface: float | None = None,
    temperature_surface: float | None = None,
) -> tuple[pd.DataFrame, dict]:
    #TODO: docstring
    #TODO: merge with generate_meteo_profiles
    # get T and P
    info = {}
    info["radiosonde_datetime"] = date.strftime(
        "%Y-%m-%dT%H-%M-%S"
    )  # FIXME: change if other methods are implemented.

    if meteorology_source == "ecmwf":  # TODO: implement other methods
        try:
            meteo_profiles = get_ecmwf_temperature_pressure(date, heights=range)
        except Exception:
            meteorology_source = "standard_atmosphere"
            raise Warning(
                "ECMWF data not available. Use meteorology_source = 'standard_atmosphere' instead."
            )
        info["radiosonde_wmo_id"] = "ecmwf"
        info["radiosonde_location"] = "Granada"
        info["radiosonde_source"] = "ECMWF"

    if meteorology_source == "standard_atmosphere":
        meteo_profiles = generate_meteo_profiles(
            range, pressure=pressure_surface, temperature=temperature_surface
        )
        info["radiosonde_location"] = "Granada"
        if pressure_surface is not None and temperature_surface is not None:        
            info["radiosonde_wmo_id"] = "scaled standard atmosphere"
            info["radiosonde_source"] = "scaled standard atmosphere"
        else:
            info["radiosonde_wmo_id"] = "standard atmosphere"
            info["radiosonde_source"] = "standard atmosphere"
        # FIXME: raise Warning( "ECMWF data not available. Using scaled standard atmosphere with (T, P) = (25, 938)" )
    else:
        raise ValueError(
            "only ecmwf and (scaled) standard_atmosphere methods are currently implemented."
        )
    return meteo_profiles, info



from datetime import datetime
from pathlib import Path
from pdb import set_trace
from typing import List, Literal, Tuple

import matplotlib
import numpy as np
import xarray as xr
from loguru import logger
import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.dates import DateFormatter
from gfatpy.lidar.utils.file_manager import filename2info

from gfatpy.lidar.preprocessing.lidar_preprocessing import preprocess
from gfatpy.lidar.utils.utils import LIDAR_INFO, signal_to_rcs
from gfatpy.utils import plot
from gfatpy.utils.io import read_yaml_from_info
from gfatpy.utils.utils import numpy_to_datetime

BoundsType = tuple[float, float] | Literal["auto", "limits"]


def get_norm(
    rcs: np.ndarray,
    scale_bounds: BoundsType,
    color_resolution: int = 128,
) -> matplotlib.colors.BoundaryNorm:
    """Get the color normalization and bounds for the colorbar.

    Args:

        - rcs (np.ndarray): Range corrected signal.
        - scale_bounds (BoundsType): Bounds for the colorbar.
        - color_resolution (int, optional): Colorbar resolution. Defaults to 128.

    Returns:

        - matplotlib.colors.BoundaryNorm: Color normalization.
    """
    match scale_bounds:
        case "auto":
            bounds = np.linspace(0, rcs.max() * 0.6, color_resolution)
        case "limits":
            bounds = np.linspace(rcs.min(), rcs.max(), color_resolution)
        case _:
            bounds = np.linspace(*scale_bounds, color_resolution)

    norm = matplotlib.colors.BoundaryNorm(boundaries=bounds, ncolors=2**8, clip=False)
    logger.debug(f"Color bounds min - max: {bounds.min()} - {bounds.max()}")
    return norm, bounds


def apply_gap_size(ax: matplotlib.axes.Axes, data_array) -> None:  # type: ignore
    """Apply gap size to the x-axis of the plot.

    Args:
        ax (matplotlib.axes.Axes): Axes object.
        data_array (_type_): Data array.
    """
    diff = data_array.time[1:].values - data_array.time[0:-1].values
    gap_size = 2 * int(
        np.ceil(
            np.median(np.median(diff).astype("timedelta64[s]").astype("float") / 60)
        )
    )
    ax.xaxis.set_major_formatter(DateFormatter("%H:%M"))  # Formato de hora: minutos

    plot.gapsizer(
        ax,
        data_array.time.values.astype("M8[ms]").astype("O"),
        data_array.range.values,
        gap_size,
        "#c7c7c7",
    )


def quicklook_xarray(
    data_array: xr.DataArray,
    lidar_name: str,
    location: str = "Granada",
    is_rcs: bool = True,
    scale_bounds: BoundsType = "auto",
    colormap: str | matplotlib.colors.Colormap = "jet",
    ylims: Tuple[float, float] = (0.0, 14000.0),
) -> Tuple[Figure, Axes]:
    """Plot a quicklook of a lidar signal.

    Args:

        - data_array (xr.DataArray): Lidar signal or rcs from `lidar.preprocess`
        - is_rcs (bool, optional): To indicate if the data is RCS or not. Defaults to True.
        - scale_bounds (BoundsType, optional): scale bounds for the colorbar. Defaults to "auto".
        - colormap (str | matplotlib.colors.Colormap, optional): colormap of the colorbar. Defaults to "jet".

    Returns:

        - tuple[Figure, Axes]: Figure and Axes objects
    """

    # Convert cmap str to matplotlib colormap
    if isinstance(colormap, str):
        colormap = plt.get_cmap(colormap)
    cmap = colormap.copy()
    cmap.set_extremes(over="white")

    # Define channel
    channel = data_array.name.split("_")[1]

    try:
        if is_rcs:
            rcs = data_array.values
        else:
            rcs = signal_to_rcs(data_array.values, data_array.range.values)

        fig, ax = plt.subplots(figsize=(15, 5))
        norm, bounds = get_norm(rcs, scale_bounds)

        q = ax.pcolormesh(
            data_array.time,
            data_array.range / 1000.0,
            rcs.T,
            cmap=cmap,
            norm=norm,
        )

        ax.set_xlabel(r"Time, $[UTC]$")
        ax.set_ylabel(r"Height, $[km, \, agl]$")
        ax.set_ylim(ylims[0] / 1000.0, ylims[1] / 1000.0)

        plot.watermark(ax, zoom=0.6, alpha=0.6)

        datestr = datetime.strftime(
            numpy_to_datetime(data_array.time[1].values), "%Y-%m-%d"
        )
        ax.set_title(f"{location} | {lidar_name.upper()} | {datestr}")

        apply_gap_size(ax, data_array=data_array)

        # Agregar la magnitud y unidades a la barra de color
        magnitud = f"RCS@{channel[:-3]}nm"
        unidades = "a.u."
        ticks = np.linspace(bounds.min(), bounds.max(), 9)
        colorbar = fig.colorbar(q, ticks=ticks, extend="max")

        # Crear el texto de magnitud y unidades
        text = f"{magnitud} [{unidades}]"

        # Calcular la mitad de la altura de la barra de color
        mid_height = (colorbar.ax.get_ylim()[0] + colorbar.ax.get_ylim()[1]) / 2

        # Añadir el texto a la barra de color girado 90 grados y 3 cm a la derecha de los valores
        x_position = 1.2  # Cambiar la coordenada x según sea necesario
        y_position = mid_height
        x_offset = 4  # 3 cm a la derecha (puede ajustarse)
        colorbar.ax.text(
            x_position + x_offset,
            y_position,
            text,
            rotation=90,
            va="center",
            ha="left",
            fontsize=12,
        )

        q.cmap.set_over("white")  # type: ignore
    except Exception as e:
        raise e
    return fig, ax


def quicklook_dataset(
    dataset: xr.Dataset,
    channel: str,
    /,
    scale_bounds: BoundsType = "auto",
    colormap: str | matplotlib.colors.Colormap = "jet",
    ylims: Tuple[float, float] = (0.0, 14000.0),
) -> Tuple[Figure, Axes]:
    """Plot a quicklook of a lidar signal.

    Args:

        - data_array (xr.DataArray): Lidar signal or rcs from `lidar.preprocess`
        - is_rcs (bool, optional): To indicate if the data is RCS or not. Defaults to True.
        - scale_bounds (BoundsType, optional): scale bounds for the colorbar. Defaults to "auto".
        - colormap (str | matplotlib.colors.Colormap, optional): colormap of the colorbar. Defaults to "jet".

    Returns:

        - tuple[Figure, Axes]: Figure and Axes objects
    """

    # Convert cmap str to matplotlib colormap
    if isinstance(colormap, str):
        colormap = plt.get_cmap(colormap)
    cmap = colormap.copy()
    cmap.set_extremes(over="white")

    data_array = dataset[f"signal_{channel}"]

    # Get lidar name from dataset
    lidar_name = dataset.attrs["location"].lower()

    quicklook_xarray(
        data_array,
        lidar_name=lidar_name,
        location="Granada",
        is_rcs=False,
        scale_bounds=scale_bounds,
        colormap=colormap,
        ylims=ylims,
    )

    try:
        rcs = signal_to_rcs(data_array.values, data_array.range.values)

        fig, ax = plt.subplots(figsize=(15, 5))
        norm, bounds = get_norm(rcs, scale_bounds)

        q = ax.pcolormesh(
            data_array.time,
            data_array.range / 1000.0,
            rcs.T,
            cmap=cmap,
            norm=norm,
        )

        ax.set_xlabel(r"Time, $[UTC]$")
        ax.set_ylabel(r"Height, $[km, \, agl]$")
        ax.set_ylim(ylims[0] / 1000.0, ylims[1] / 1000.0)

        datestr = datetime.strftime(
            numpy_to_datetime(data_array.time[1].values), "%Y%m%d"
        )
        plot.title1(f"{data_array.name} | {datestr}", 2)

        plot.watermark(ax, zoom=0.6, alpha=0.6)
        apply_gap_size(ax, data_array=data_array)

        # Agregar la magnitud y unidades a la barra de color
        magnitud = "RCS"
        unidades = "a.u."
        ticks = np.linspace(bounds.min(), bounds.max(), 9)
        colorbar = fig.colorbar(q, ticks=ticks, extend="max")

        # Crear el texto de magnitud y unidades
        text = f"{magnitud} [{unidades}]"

        # Calcular la mitad de la altura de la barra de color
        mid_height = (colorbar.ax.get_ylim()[0] + colorbar.ax.get_ylim()[1]) / 2

        # Añadir el texto a la barra de color girado 90 grados y 3 cm a la derecha de los valores
        x_position = 1.2  # Cambiar la coordenada x según sea necesario
        y_position = mid_height
        x_offset = 4  # 3 cm a la derecha (puede ajustarse)
        colorbar.ax.text(
            x_position + x_offset,
            y_position,
            text,
            rotation=90,
            va="center",
            ha="left",
            fontsize=12,
        )

        q.cmap.set_over("white")  # type: ignore
    except Exception as e:
        raise e
    return fig, ax


def quicklook_from_file(
    filepath: Path | str,
    channels: List[str],
    output_dir: Path | str | None = None,
    scale_bounds: BoundsType = "auto",
    **kwargs,
):
    # Define output directory
    if output_dir is None:
        output_dir = Path.cwd()
    elif isinstance(output_dir, str):
        output_dir = Path(output_dir)
    else:
        if not output_dir.exists():
            raise ValueError(f"Output directory {output_dir} does not exist.")

    # Check filepath
    if isinstance(filepath, str):
        filepath = Path(filepath)

    if not filepath.exists():
        raise FileNotFoundError(f"File {filepath} does not exist.")

    # Get lidar name from filepath
    nickname, _, _, _, _, target_date = filename2info(filepath.name)
    lidar_name = LIDAR_INFO["metadata"]["nick2name"][nickname]

    # Get INFO
    INFO = read_yaml_from_info(nickname, target_date)

    if kwargs.get("apply_ov", False):
        channels2load = channels.copy()
        for channel_ in channels2load:
            if "f" in channel_:
                overlap_channel_ = INFO["overlap_channels"][channel_]
                if overlap_channel_ not in channels2load:
                    channels2load.append(overlap_channel_)
    else:
        channels2load = None

    lidar = preprocess(
        filepath,
        channels=channels2load,
        apply_bg=kwargs.get("apply_bg", True),
        apply_dc=kwargs.get("apply_dc", False),
        apply_dt=kwargs.get("apply_dt", False),
        apply_bz=kwargs.get("apply_bz", True),
        apply_ov=kwargs.get("apply_ov", False),
    )

    if 'ghk_dir' in kwargs and 'depo_calib_dir' in kwargs:
        lidar.add_depolarization_products(lidar, ghk_dir=kwargs['ghk_dir'], depo_calib_dir=kwargs['depo_calib_dir'])    

    # Get location from dataset
    location = lidar.attrs["location"]

    # Get date string
    datestr = datetime.strftime(numpy_to_datetime(lidar.time[1].values), "%Y%m%d")

    for channel_ in channels:
        fig, _ = quicklook_xarray(
            lidar[f"signal_{channel_}"],
            lidar_name=lidar_name,
            location=location,
            is_rcs=False,
            scale_bounds=scale_bounds,
        )
        fig.savefig(
            output_dir / f"quicklook_{nickname}_{channel_}_{datestr}.png",
            dpi=600,
            bbox_inches="tight",
        )
        plt.close(fig)
    lidar.close()


# TODO: this should be the default implemetation rather than filelist argument



import os
import itertools
import time
import numpy as np
import pandas as pd
import xarray as xr
import matplotlib as mpl
import matplotlib.dates as mdates
from matplotlib.collections import LineCollection
import datetime as dt
from cycler import cycler
import matplotlib.pyplot as plt
from matplotlib.ticker import MultipleLocator
import matplotlib.patches as mpatches
from gfatpy import utils

from gfatpy.lidar.preprocessing.lidar_preprocessing import preprocess
from gfatpy.lidar.preprocessing.lidar_preprocessing_tools import (
    apply_dead_time_correction,
)
from gfatpy.lidar.preprocessing.gluing_proportional import gluing
from gfatpy.utils.utils import residuals

# lidar = importlib.import_module("%s.lidar" % config.gfat_name)
# utils = importlib.import_module("%s.utils" % config.gfat_name)
# solar = importlib.import_module("%s.solar" % config.gfat_name)
plt.ion()


def plot_hist_num_measurements(year=None):
    """

    :return:
    """

    if year is not None:
        year_ini = year
        year_end = year
    else:
        year_ini = 2018
        year_end = 2020

    date_ini = dt.datetime(year_ini, 1, 1)
    date_end = dt.datetime(year_end, 12, 31)
    dates_ls = pd.date_range(date_ini, date_end)
    num_meas = np.zeros(len(dates_ls))
    for i, idate in enumerate(dates_ls):
        print(idate)
        i_fn = os.path.join(
            ".",
            "dead_time",
            "dead_time_%s_532p_ngt_scc_1.nc" % idate.strftime("%Y%m%d"),
        )
        if os.path.isfile(i_fn):
            try:
                ds = xr.open_dataset(i_fn)
                # corrijo un fallo de seleccion de medidas
                times_a = ds.times.values
                sun = solar.SUN(times_a, -3.61, 37.16, elev=680)
                csza = sun.get_csza()
                idx_ilu = np.where(csza < -0.01)
                times_ilu = times_a[idx_ilu]
                ds = ds.sel(times=times_ilu)
                num_meas[i] = ds.dims["times"]
                ds.close()
            except:
                print("%s: no data" % idate)
    df = pd.DataFrame.from_dict({"times": dates_ls, "num": num_meas})
    plt.close("all")
    f, ax = plt.subplots(figsize=(8.5, 3.75), constrained_layout=True)
    ax.bar(df.times, df.num)
    ax.xaxis.set_major_formatter(mdates.DateFormatter("%b"))
    ax.axes.set_xlim("2019-01-01", "2019-12-31")
    ax.axes.set_xlabel("Date")
    ax.axes.set_ylabel("num. of measurements")
    if year is not None:
        ax.set_title("MULHACEN: measurements along %s" % year)
        plt.savefig("lidar_measurements_%s.png" % (year,), dpi=300)
    else:
        ax.set_title("MULHACEN: measurements along %s-%s" % (year_ini, year_end))
        plt.savefig("lidar_measurements_%s_%s.png" % (year_ini, year_end), dpi=300)

    return df


def plot_J_lines(fn=None, ds=None):
    """ """
    if fn is not None:
        ds = xr.open_dataset(fn)
    dx = ds.isel(times=10)

    colors = mpl.cm.jet(np.linspace(0, 1, ds.dims["steps"]))
    custom_cycler = cycler(color=colors)
    cm = mpl.cm.get_cmap("jet", ds.dims["steps"])

    plt.close("all")
    f, ax = plt.subplots(constrained_layout=True)
    ax.set_prop_cycle(custom_cycler)
    lines = [
        np.column_stack([dx.tau.values, np.log(dx.J).values[:, i]])
        for i in range(dx.dims["steps"])
    ]
    lc = LineCollection(lines, cmap=cm, lw=3)
    lc.set_array(dx.range_glue_extended_thick[0, :].values)
    ax.add_collection(lc)
    axcb = f.colorbar(lc, ax=ax)

    x_label = r"dead time [ns]"
    y_label = r"residual, log(J)"
    hue_label = r"region thickness [m]"
    plot_title = r"J vs $\tau$ vs linear response region thickness"

    ax.axes.set_ylim(np.log(dx.J).min(skipna=True), np.log(dx.J).max(skipna=True))
    ax.axes.set_xlim(dx.tau.min(skipna=True) - 0.1, dx.tau.max(skipna=True) + 0.1)

    label_size = 12
    tick_label_size = 11

    ax.tick_params(axis="x", labelsize=tick_label_size)
    ax.tick_params(axis="y", labelsize=tick_label_size)
    ax.set_xlabel(x_label, fontsize=label_size)
    ax.set_ylabel(y_label, fontsize=label_size)
    ax.set_title(plot_title)
    axcb.set_label(hue_label, fontsize=label_size)
    axcb.ax.tick_params(labelsize=tick_label_size)

    plt.savefig("J_lines.png", dpi=300)


def plot_J_contour(fn=None, ds=None):
    """ """
    if fn is not None:
        ds = xr.open_dataset(fn)
    dx = ds.isel(times=10)

    plt.close("all")
    f, ax = plt.subplots(constrained_layout=True)
    cs = ax.contourf(
        np.tile(dx.tau, [dx.dims["steps"], 1]).T,
        dx.range_glue_extended_thick,
        np.log(dx.J),
        cmap="jet",
    )
    axcb = f.colorbar(cs)

    x_label = r"dead time [ns]"
    y_label = r"region thickness [m]"
    hue_label = r"residual, log(J)"
    plot_title = r"J vs $\tau$ vs linear response region thickness"

    ax.axes.set_xlim(dx.tau.min(skipna=True) - 0.1, dx.tau.max(skipna=True) + 0.1)

    label_size = 12
    tick_label_size = 11

    ax.tick_params(axis="x", labelsize=tick_label_size)
    ax.tick_params(axis="y", labelsize=tick_label_size)
    ax.set_xlabel(x_label, fontsize=label_size)
    ax.set_ylabel(y_label, fontsize=label_size)
    ax.set_title(plot_title)
    axcb.set_label(hue_label, fontsize=label_size)
    axcb.ax.tick_params(labelsize=tick_label_size)

    plt.savefig("J_contour.png", dpi=300)


def plot_scatter_dead_time(fn=None, ds=None):
    """
    SCATTER ESTIMATED DEAD TIME AS DEPENDENT ON REGION THICKNESS
    """

    if fn is not None:
        ds = xr.open_dataset(fn)

    date_ini = ds.times[0].values
    date_end = ds.times[-1].values
    date_ini = utils.numpy_to_datetime(date_ini)
    date_end = utils.numpy_to_datetime(date_end)
    delta_t = (date_end - date_ini).seconds / 3600
    plot_title = "%s: %.1f hours. %i profiles" % (
        date_ini.strftime("%Y%m%d"),
        delta_t,
        ds.dims["times"],
    )
    x_label = r"dead time [ns]"
    y_label = r"region thickness [m]"
    hue_label = r"r Pearson (Glued - Photon-counting)"

    label_size = 12
    tick_label_size = 11

    # height limits
    # h_limits = np.arange(0, 7500, 500)
    color_map = "jet"
    hue_limits = (0.95, 1)
    hue_step = 0.005
    num_colors = round((hue_limits[1] - hue_limits[0]) / hue_step)

    f, ax = plt.subplots(constrained_layout=True)
    # plot
    for i in range(ds.dims["times"]):
        cs = ds.isel(times=i).plot.scatter(
            ax=ax,
            x="tau_min",
            y="range_glue_extended_thick",
            hue="rvalue_gl_pc",
            cmap=mpl.cm.get_cmap(color_map, num_colors),
            vmin=hue_limits[0],
            vmax=hue_limits[1],
            extend="min",
            add_guide=False,
        )  # type: ignore
    ax.tick_params(axis="x", labelsize=tick_label_size)
    ax.tick_params(axis="y", labelsize=tick_label_size)
    ax.set_xlabel(x_label, fontsize=label_size)
    ax.set_ylabel(y_label, fontsize=label_size)
    ax.set_title(plot_title)
    # colorbar
    cb = plt.colorbar(cs)
    # cb.ax.locator_params(nbins=num_colors)
    # cb.ax.set_title('hola')
    cb.set_label(hue_label, fontsize=label_size)
    cb.ax.tick_params(labelsize=tick_label_size)

    plt.savefig("dead_time_distribution_region_thickness.png", dpi=300)

    ds.close()


def plot_dead_time_hist(fn=None, ds=None, label=None):
    """ """
    if fn is not None:
        ds = xr.open_dataset(fn)

    date_ini = ds.times[0].values
    date_end = ds.times[-1].values
    date_ini = utils.numpy_to_datetime(date_ini)
    date_end = utils.numpy_to_datetime(date_end)
    delta_t = (date_end - date_ini).seconds / 3600
    plot_title = "%s: %.1f hours. %i profiles" % (
        date_ini.strftime("%Y%m%d"),
        delta_t,
        ds.dims["times"],
    )
    label_size = 12
    tick_label_size = 11

    f, ax = plt.subplots(constrained_layout=True)
    ds.tau_best_avg.plot.hist(ax=ax, bins=ds.tau, color="royalblue")
    ax.xaxis.set_major_locator(MultipleLocator(1))
    ax.xaxis.set_minor_locator(MultipleLocator(0.5))
    ax.axes.set_xlabel("Dead Time [ns]", fontsize=label_size)
    ax.axes.set_ylabel("Counts", fontsize=label_size)
    ax.set_title(plot_title)  # , fontsize=16)
    ax.tick_params(axis="x", labelsize=tick_label_size)
    ax.tick_params(axis="y", labelsize=tick_label_size)
    # l1 = mpatches.Patch(color='royalblue') #, label="day")
    # l1_t = mpatches.Patch(color='w', label="\n".join((r"$\mu = %.1f$" % (ds_day.tau_min.mean(), ), r"$\sigma = %.1f$" % (ds_day.tau_min.std(),))))
    l1_t = mpatches.Patch(
        color="royalblue",
        label="\n".join(
            (
                r"$\mu = %.1f$" % (ds.tau_best_avg.mean(),),
                r"$\sigma = %.1f$" % (ds.tau_best_avg.std(),),
            )
        ),
    )
    ax.legend(handles=[l1_t], fontsize=label_size)

    if label is not None:
        plt_fn = "dead_time_hist_%s.png" % label
    else:
        plt_fn = "dead_time_hist.png"
    plt.savefig(plt_fn, dpi=300)

    ds.close()


def plot_dead_time_ts(year=None, df=None):
    """ """

    if df is None:
        if year is not None:
            year_ini = year
            year_end = year
        else:
            year_ini = 2018
            year_end = 2020

        date_ini = dt.datetime(year_ini, 1, 1)
        date_end = dt.datetime(year_end, 12, 31)
        dates_ls = pd.date_range(date_ini, date_end)
        tau_532p_avg = np.zeros(len(dates_ls)) * np.nan
        tau_532p_std = np.zeros(len(dates_ls)) * np.nan
        tau_532c_avg = np.zeros(len(dates_ls)) * np.nan
        tau_532c_std = np.zeros(len(dates_ls)) * np.nan
        tau_355_avg = np.zeros(len(dates_ls)) * np.nan
        tau_355_std = np.zeros(len(dates_ls)) * np.nan
        for i, idate in enumerate(dates_ls):
            print(idate)
            i_fn = os.path.join(
                ".",
                "dead_time",
                "dead_time_%s_532p_ngt_scc_1.nc" % idate.strftime("%Y%m%d"),
            )
            # if "20191010" in i_fn:
            #    pdb.set_trace()
            if os.path.isfile(i_fn):
                try:
                    ds = xr.open_dataset(i_fn)
                    tau_532p_avg[i] = ds.tau_best_avg.mean(skipna=True)
                    tau_532p_std[i] = ds.tau_best_avg.std(skipna=True)
                    ds.close()
                except:
                    print("%s: no data for 532p" % idate)
            i_fn = os.path.join(
                ".",
                "dead_time",
                "dead_time_%s_532c_ngt_scc_1.nc" % idate.strftime("%Y%m%d"),
            )
            if os.path.isfile(i_fn):
                try:
                    ds = xr.open_dataset(i_fn)
                    tau_532c_avg[i] = ds.tau_best_avg.mean(skipna=True)
                    tau_532c_std[i] = ds.tau_best_avg.std(skipna=True)
                    ds.close()
                except:
                    print("%s: no data for 532c" % idate)
            i_fn = os.path.join(
                ".",
                "dead_time",
                "dead_time_%s_355_ngt_scc_1.nc" % idate.strftime("%Y%m%d"),
            )
            if os.path.isfile(i_fn):
                try:
                    ds = xr.open_dataset(i_fn)
                    tau_355_avg[i] = ds.tau_best_avg.mean(skipna=True)
                    tau_355_std[i] = ds.tau_best_avg.std(skipna=True)
                    ds.close()
                except:
                    print("%s: no data for 355" % idate)

        df = pd.DataFrame.from_dict(
            {
                "times": dates_ls,
                "tau_532p_avg": tau_532p_avg,
                "tau_532p_std": tau_532p_std,
                "tau_532c_avg": tau_532c_avg,
                "tau_532c_std": tau_532c_std,
                "tau_355_avg": tau_355_avg,
                "tau_355_std": tau_355_std,
            }
        )
    # plot
    plt.close("all")
    f, ax = plt.subplots(figsize=(8.5, 3.75), constrained_layout=True)
    ax.plot(df.times, df.tau_532p_avg, lw=0, label="532p", color="b", marker="o")
    # ax.fill_between(df.times, df.tau_532p_avg + df.tau_532p_std, df.tau_532p_avg - df.tau_532p_std, facecolor='b', alpha=0.5)
    ax.plot(df.times, df.tau_532c_avg, lw=0, label="532c", color="r", marker="o")
    # ax.fill_between(df.times, df.tau_532c_avg + df.tau_532c_std, df.tau_532c_avg - df.tau_532c_std, facecolor='r', alpha=0.5)
    ax.plot(df.times, df.tau_355_avg, lw=0, label="355", color="k", marker="o")
    # ax.fill_between(df.times, df.tau_355_avg + df.tau_355_std, df.tau_355_avg - df.tau_355_std, facecolor='k', alpha=0.5)
    label_size = 12
    tick_label_size = 11

    ax.legend(fontsize=label_size)
    ax.xaxis.set_major_formatter(mdates.DateFormatter("%b"))
    ax.axes.set_xlim("2019-01-01", "2019-12-31")
    ax.tick_params(axis="x", labelsize=tick_label_size)
    ax.tick_params(axis="y", labelsize=tick_label_size)
    ax.set_xlabel("Date", fontsize=label_size)
    ax.set_ylabel("dead time [ns]", fontsize=label_size)
    if year is not None:
        ax.set_title("MULHACEN: dead time estimates. Year %s" % year)
        plt.savefig("dead_time_ts_%s.png" % (year,), dpi=300)
    else:
        ax.set_title(
            "MULHACEN: dead time estimates. Period %s-%s" % (year_ini, year_end)
        )
        plt.savefig("dead_time_ts_%s_%s.png" % (year_ini, year_end), dpi=300)

    return df


#        f, ax = plt.subplots(constrained_layout=True)
#        cs = ax.contourf(np.tile(tau, [n_steps, 1]).T, range_glue_ext_thick[0, :, :],
#                         J[0, :, :], levels=np.arange(0, 0.1, 0.005), cmap='jet')
#        plt.colorbar(cs)
#
#        # Plot rvalue vs thickness vs tau
#        f, ax = plt.subplots(constrained_layout=True)
#        cs = ax.contourf(np.tile(tau, [n_steps, 1]).T, range_glue_ext_thick[0, :, :],
#                         rvalue[0, :, :], levels=np.arange(0.8, 1.01, 0.01), cmap='jet')
#        plt.colorbar(cs)
#
#        # Plot J vs tau for every step
#        colors = mpl.cm.jet(np.linspace(0, 1, n_steps))
#        custom_cycler = cycler(color=colors)
#        f, ax = plt.subplots(constrained_layout=True)
#        ax.set_prop_cycle(custom_cycler)
#        ax.plot(tau, J[0, :, :])
#
#        # Plot tau_min vs thickness with color (min_range, rvalue of tau_min, num of profile)
#        x = []
#        y = []
#        c_thick = []
#        c_min_rvalue = []
#        c_profile = []
#        for i in range(n_t):
#            for j in range(n_tau):
#                if len(x) == 0:
#                    x = tau_min[i, :]
#                    y = range_glue_ext_min[i, j, :]
#                    c_thick = range_glue_ext_thick[i, j, :]
#                    c_min_rvalue = r_min[i, :]
#                    c_profile = np.tile(i, n_steps)
#                else:
#                    x = np.append(x, tau_min[i, :])
#                    y = np.append(y, range_glue_ext_min[i, j, :])
#                    c_thick = np.append(c_thick, range_glue_ext_thick[i, j, :])
#                    c_min_rvalue = np.append(c_min_rvalue, r_min[i, :])
#                    c_profile = np.append(c_profile, np.tile(i, n_steps))
#        colors_min_range = mpl.cm.get_cmap("jet", round((2000-0)/200))
#        f, ax = plt.subplots(constrained_layout=True)
#        cs = ax.scatter(x, y, marker='+', c=c_thick, cmap=colors_min_range, vmin=0, vmax=2000)
#        cb = plt.colorbar(cs)
#        cb.ax.locator_params(nbins=round((2000-0)/200))
#
#        colors_min_rvalue = mpl.cm.get_cmap("jet", round((1-0.9)/0.01))
#        f, ax = plt.subplots(constrained_layout=True)
#        cs = ax.scatter(x, y, marker='+', c=c_min_rvalue, cmap=colors_min_rvalue, vmin=0.9, vmax=1)
#        plt.colorbar(cs)
#
#        colors_profiles = mpl.cm.get_cmap("jet", n_t)
#        f, ax = plt.subplots(constrained_layout=True)
#        cs = ax.scatter(x, y, marker='+', c=c_profile, cmap=colors_profiles, vmin=-0.5, vmax=n_t-0.5)
#        plt.colorbar(cs)


def estimate_dead_time_old(
    date_str, channel, don, time_range=None, gluing_type="scc", expand_bins=True
):
    """ """

    # Group results
    if don == 1:
        don_str = "day"
    elif don == 2:
        don_str = "ngt"
    elif don == 0:
        don_str = "all"
    else:
        don_str = "_".join([str(t) for t in time_range])

    try:
        t_start = time.time()
        print(
            "Estimate Dead Time for %s. channel=%s. illum=%i" % (date_str, channel, don)
        )
        # get raw data
        (
            an,
            pc,
            an_dc,
            ranges,
            times,
            bg_range,
            rs_range,
            idx_bg_range,
            idx_rs_range,
            rs_ds,
            dc_ds,
            an_ch,
            pc_ch,
            wv,
            pol,
        ) = lidar_read.get_prepared_data(
            date_str, channel, don, 15000, time_range=time_range
        )
        # analog channel additional info
        channel_id = np.logical_and.reduce(
            (
                rs_ds.wavelength == wv,
                rs_ds.polarization == pol,
                rs_ds.detection_mode == 0,
            )
        )
        adc_range = 1000 * rs_ds.adc_range.values[channel_id][0]
        adc_bits = rs_ds.adc_bits.values[channel_id][0]
        n_res = 1

        # test 10 profiles
        # n_p_test = 5
        # times = times[:n_p_test]
        # an = an[:n_p_test, :]
        # pc = pc[:n_p_test, :]

        # taus [in nanoseconds]
        tau_step = 0.1  # 0.05
        tau = np.arange(0, 10 + tau_step, tau_step)

        # expand glue region
        bin_step = 10
        bin_max = 150
        exp_glue = np.arange(0, bin_max + bin_step, bin_step) * 7.5

        # Previous info for preprocessing signals
        bz_an = lidar.get_bin_zero(wv, pol, 0, ref_time=dt.datetime(2020, 5, 28))
        bz_pc = lidar.get_bin_zero(wv, pol, 1, ref_time=dt.datetime(2020, 5, 28))
        bg_an = lidar.estimate_raw_background(an, ranges, bg_range=bg_range)
        bg_pc = lidar.estimate_raw_background(pc, ranges, bg_range=bg_range)
        dc, dc_bg = lidar.preprocessing_dc_signal(an_dc, ranges, bz_an, bg_range)

        # re-dimension signals with range limits
        an = an[:, idx_rs_range]
        pc = pc[:, idx_rs_range]
        dc = dc[idx_rs_range]
        ranges = ranges[idx_rs_range]

        # dimensions
        n_t = len(times)
        n_ranges = len(ranges)
        n_tau = len(tau)
        n_steps = len(exp_glue)

        # Define variables for output
        tau_best_avg = np.zeros(n_t)
        tau_best_std = np.zeros(n_t)
        thickness_best_avg = np.zeros(n_t)
        thickness_best_std = np.zeros(n_t)
        tau_min = np.zeros((n_t, n_steps)) * np.nan
        r_min = np.zeros((n_t, n_steps)) * np.nan
        r_min2 = np.zeros((n_t, n_steps)) * np.nan
        thickness_minJ = np.zeros((n_t, n_steps)) * np.nan
        J = np.zeros((n_t, n_tau, n_steps)) * np.nan
        range_glue_first_min = np.zeros((n_t, n_tau)) * np.nan
        range_glue_first_max = np.zeros((n_t, n_tau)) * np.nan
        range_glue_min = np.zeros((n_t, n_tau)) * np.nan
        range_glue_max = np.zeros((n_t, n_tau)) * np.nan
        an_pp = np.zeros((n_t, n_ranges)) * np.nan
        pc_pp = np.zeros((n_t, n_ranges, n_tau)) * np.nan
        gl_pp = np.zeros((n_t, n_ranges, n_tau)) * np.nan
        rvalue = np.zeros((n_t, n_tau, n_steps)) * np.nan
        rvalue2 = np.zeros((n_t, n_tau, n_steps)) * np.nan
        range_glue_ext_min = np.zeros((n_t, n_tau, n_steps)) * np.nan
        range_glue_ext_max = np.zeros((n_t, n_tau, n_steps)) * np.nan
        range_glue_ext_thick = np.zeros((n_t, n_tau, n_steps)) * np.nan

        # Loop over signals
        for i in range(n_t):
            # preprocess analog
            an_i = lidar.preprocessing_analog_signal(
                an[i, :], bz_an, bg_an[i], dc, dc_bg
            )
            an_pp[i, :] = an_i

            # Loop over taus
            for j in range(n_tau):
                # preprocess photoncounting
                pc_ij = lidar.preprocessing_photoncounting_signal(
                    pc[i, :], tau[j], bz_pc, bg_pc[i]
                )
                pc_pp[i, :, j] = pc_ij

                # gluing
                if gluing_type == "scc":  # D'Amico
                    gl_ij, c_an, c_ph, idxs, glued = lidar.gluing(
                        an_i,
                        pc_ij,
                        ranges,
                        adc_range,
                        adc_bits,
                        n_res=n_res,
                        pc_threshold=20,
                        correlation_threshold=0.9,
                        range_threshold=(1000, 5000),
                        min_points=15,
                        slope_threshold=2,
                        stability_threshold=1,
                        step=5,
                        use_photon_as_reference=True,
                    )
                    if glued:
                        # ranges of gluing region
                        range_glue_first_min[i, j] = ranges[idxs[0]]
                        range_glue_first_max[i, j] = ranges[idxs[1]]
                        range_glue_min[i, j] = ranges[idxs[4]]
                        range_glue_max[i, j] = ranges[idxs[5]]
                        gl_pp[i, :, j] = gl_ij
                else:  # method dbp
                    print("Method dbp to be implemented")

            # Cost Function: para cada perfil uso el mismo rango para todos los taus
            min_glue_i = np.nanmax(range_glue_min[i, :])
            max_glue_i = np.nanmin(range_glue_max[i, :])
            for j in range(n_tau):
                for k in range(n_steps):
                    # anchura de region
                    min_glue = min_glue_i - exp_glue[k]
                    max_glue = max_glue_i
                    idx_glue = np.logical_and(ranges >= min_glue, ranges <= max_glue)
                    # rangos
                    range_glue_ext_min[i, j, k] = min_glue
                    range_glue_ext_max[i, j, k] = max_glue
                    range_glue_ext_thick[i, j, k] = max_glue - min_glue

                    # correlación AN vs PC en la region
                    rvalue[i, j, k] = np.corrcoef(
                        an_pp[i, idx_glue], pc_pp[i, idx_glue, j]
                    )[0, 1]
                    rvalue2[i, j, k] = np.corrcoef(
                        gl_pp[i, idx_glue, j], pc_pp[i, idx_glue, j]
                    )[0, 1]
                    if np.logical_and(
                        ~np.isnan(rvalue2[i, j, k]), rvalue2[i, j, k] >= 0.9
                    ):
                        if rvalue[i, j, k] >= 0.9:
                            J[i, j, k] = utils.residuals(
                                pc_pp[i, idx_glue, j], gl_pp[i, idx_glue, j]
                            )
            # Tau that Minimizes Cost Function and associated rvalues for an_pc and gl_pc
            for k in range(n_steps):
                jj = J[i, :, k]
                rr = rvalue[i, :, k]
                rr2 = rvalue2[i, :, k]
                th = range_glue_ext_thick[i, :, k]
                try:
                    if ~(np.isnan(jj).any()):
                        idx_Jmin = np.nanargmin(jj)
                        tau_min[i, k] = tau[idx_Jmin]
                        r_min[i, k] = rr[idx_Jmin]
                        r_min2[i, k] = rr2[idx_Jmin]
                        thickness_minJ[i, k] = th[idx_Jmin]
                except Exception as e:
                    print(str(e))
                    print(
                        "Tau min not estimated for profile %i-th, step %i-th" % (i, k)
                    )

            # CRITERIUM TO ESTIMATE DEAD TIME
            # for every thickness, a best tau is estimated above.
            # So, we have a list of best taus. from them, we derive the best
            # of the best, considering the correlation index gl_pc
            try:
                idx_best_tau = r_min2[i, :] >= (
                    np.nanmax(r_min2[i, :]) - np.nanstd(r_min2[i, :])
                )
                tau_best_avg[i] = np.nanmean(tau_min[i, idx_best_tau])
                tau_best_std[i] = np.nanstd(tau_min[i, idx_best_tau])
                thickness_best_avg[i] = np.nanmean(thickness_minJ[i, idx_best_tau])
                thickness_best_std[i] = np.nanstd(thickness_minJ[i, idx_best_tau])
            except Exception as e:
                print(str(e))
                print("Best Tau not estimated for profile %i-th" % i)

        # save results in netcdf
        ds = xr.Dataset(
            {
                "J": (["times", "tau", "steps"], J),
                "tau_min": (["times", "steps"], tau_min),
                "tau_best_avg": (["times"], tau_best_avg),
                "tau_best_std": (["times"], tau_best_std),
                "thickness_best_avg": (["times"], thickness_best_avg),
                "thickness_best_std": (["times"], thickness_best_std),
                "range_glue_extended_min": (
                    ["times", "tau", "steps"],
                    range_glue_ext_min,
                ),
                "range_glue_extended_max": (
                    ["times", "tau", "steps"],
                    range_glue_ext_max,
                ),
                "range_glue_extended_thick": (
                    ["times", "tau", "steps"],
                    range_glue_ext_thick,
                ),
                "rvalue_an_pc": (["times", "tau", "steps"], rvalue),
                "rvalue_gl_pc": (["times", "tau", "steps"], rvalue2),
            },
            coords={"times": times, "tau": tau, "steps": exp_glue, "ranges": ranges},
        )
        fn = os.path.join(
            config.dead_time_dn,
            "dead_time_%s_%s_%s_%s_%i.nc"
            % (date_str, channel, don_str, gluing_type, expand_bins),
        )
        ds.to_netcdf(fn)
        time_elapsed = time.time() - t_start
        print(
            "%.1f s of time elapsed for %s:%s (%i profiles). channel=%s"
            % (time_elapsed, date_str, don_str, ds.dims["times"], channel)
        )

    except Exception as e:
        print("ERROR: In study_dead_time. %s" % (str(e)))
        ds = None
        fn = ""

    return ds, fn


def run_period(date_ini, date_end, expand_step=True):
    """
    Perform Dead Time Estimation for a given period

    Params:
    ------
    date_ini (str, yyyymmdd):
    date_end (str, yyyymmdd):
    expand_step (bool):
        estimate for several thicknesses of gluing region
    """

    channel = ["532p", "532c", "355"]
    don = [1, 2]
    dates_str = [x.strftime("%Y%m%d") for x in pd.date_range(date_ini, date_end)]
    inp = [
        x
        for x in itertools.product(
            dates_str, channel, don, [None], ["scc"], [expand_step]
        )
    ]
    for x in inp:
        print(x)
        estimate_dead_time(*x)


if __name__ == "__main__":

    #    # Study of Impact of Gluing Range
    #    date_str = "20190108" #"20190516" #"20200702"
    #    channel = "532p"  # "532c" "355" "532p"
    #    don = 1
    #    time_range = None
    #    #time_range = (21, 22) #None# None (10, 11)
    #    ds, fn = estimate_dead_time(date_str, channel, don, time_range=time_range)
    #    #fn = "./dead_time/dead_time_20190516_20190516_532p_21_22_scc.nc"
    #    #fn = "./dead_time/dead_time_20200702_20200702_532p_21_22_scc.nc"
    #    #fn = "./dead_time/dead_time_20200702_20200702_532p_21_22_scc_1.nc"
    #    #fn = "./dead_time/dead_time_20200702_20200702_355_21_22_scc_1.nc"
    # fn = "./dead_time/dead_time_20200702_20200702_532c_21_22_scc_1.nc"
    # ds = study_impact_of_range_selection_on_dead_time_estimation(fn)

    # Run Period
    date_ini = "20190101"
    date_end = "20201031"
    run_period(date_ini, date_end, expand_step=True)

#    # Test Time Interval
#    date_ini = "20200702" #"20190516" #"20200702"
#    date_end = "20200702" #"20190516" #"20200702"
#    channel = "532p"  # "532c" "355" "532p"
#    don = -1
#    time_range = (21, 22) #None# None (10, 11)
#    x = estimate_dead_time(date_ini, date_end, channel, don, time_range=time_range, expand_step=True)

#    # Test Estimate Dead Time Period
#    date_ini = "20200528"
#    date_end = "20200528"
#    result = run_study_dead_time(date_ini, date_end)

#    # Test Gluing
#    date_str = "20200528"
#    an, pc, an_dc, ranges, times, bg_range, rs_range, idx_bg_range, idx_rs_range, \
#    rs_ds, dc_ds, an_ch, pc_ch, wv, pol = lidar_test.get_prepared_data(
#        date_str, "532p", -1, 15000, time_range=(10, 11))
#    # adc_bits, adc_range
#    channel_id = np.logical_and.reduce((rs_ds.wavelength == wv,
#                                        rs_ds.polarization == pol,
#                                        rs_ds.detection_mode == 0))
#    adc_range = 1000*rs_ds.adc_range[channel_id].values[0]
#    adc_bits = rs_ds.adc_bits[channel_id].values[0]
#    n_res = (2**adc_bits - 1)/5000.0
#
#    # preprocess with default dead time
#    bz_an = lidar.get_bin_zero(wv, pol, 0)
#    bz_pc = lidar.get_bin_zero(wv, pol, 1)
#    bg_an = lidar.estimate_raw_background(an, ranges, bg_range=bg_range)
#    dc, dc_bg = lidar.preprocessing_dc_signal(an_dc, ranges, bz_an, bg_range)
#    an = lidar.preprocessing_analog_signal(an, bz_an, bg_an, dc, dc_bg)
#    pc = lidar.preprocessing_photoncounting_signal(pc, 3.703, bz_pc, ranges, bg_range)
#
#    gl, c_an, c_ph, idxs = lidar.gluing(an[0, :], pc[0, :], ranges, adc_range, adc_bits, n_res=n_res,
#                 pc_threshold=20, r2_threshold=0.8, range_threshold=(1000, 5000),
#                 min_points=5, slope_threshold=2, stability_threshold=1, step=5,
#                 use_photon_as_reference=True)


def estimate_dead_time(
    filepath,
    channel,
    time_range=None,
    tau_range=None,
    gluing_range=(3000.0, 4000.0),
    gluing_type="scc",
    expand_bins=True,
):

    channel_an = channel + "a"
    channel_pc = channel + "p"

    channels = [channel_an, channel_pc]

    lidar = preprocess(
        filepath,
        channels=channels,
        apply_dt=False,
        save_bg=True,
        save_dc=True,
        apply_bz=False,
    )
    if time_range is not None:
        lidar = lidar.sel(time=slice(*time_range))

    if not channel_an in lidar.channel:
        raise ValueError("channel not found.")

    if not channel_pc in lidar.channel:
        raise ValueError("channel not found.")

    # select signal
    an = lidar[f"signal_{channel_an}"].sel(range=slice(*gluing_range))
    pc = lidar[f"signal_{channel_pc}"].sel(range=slice(*gluing_range))

    # select background
    lidar[f"bg_{channel_an}"].values
    lidar[f"bg_{channel_pc}"].values

    # select dark current
    dc_an = lidar[f"dc_{channel_an}"].sel(range=slice(*gluing_range)).values

    # select range
    ranges = lidar.range.sel(range=slice(*gluing_range)).values
    times = lidar.time.values

    # analog channel additional info
    adc_range = 1000 * lidar.adc_range.sel(channel=channel_an).values.item()
    adc_bits = lidar.adc_bits.sel(channel=channel_an).values.item()

    # taus [in nanoseconds]
    if tau_range is None:
        tau_step = 1.0  # 0.05
        tau = np.arange(3, 5 + tau_step, tau_step)
    else:
        tau = np.arange(*tau_range)

    # expand glue region
    bin_step, bin_min, bin_max = 10, 75, 100
    exp_glue = np.arange(bin_min, bin_max + bin_step, bin_step) * 7.5

    # dimensions
    n_t = len(times)
    n_ranges = len(ranges)
    n_tau = len(tau)
    n_steps = len(exp_glue)

    # Define variables for output
    tau_best_avg = np.zeros(n_t)
    tau_best_std = np.zeros(n_t)
    thickness_best_avg = np.zeros(n_t)
    thickness_best_std = np.zeros(n_t)
    tau_min = np.zeros((n_t, n_steps)) * np.nan
    r_min = np.zeros((n_t, n_steps)) * np.nan
    r_min2 = np.zeros((n_t, n_steps)) * np.nan
    thickness_minJ = np.zeros((n_t, n_steps)) * np.nan
    cost_function = np.zeros((n_t, n_tau, n_steps)) * np.nan
    np.zeros((n_t, n_tau)) * np.nan
    np.zeros((n_t, n_tau)) * np.nan
    np.zeros((n_t, n_tau)) * np.nan
    np.zeros((n_t, n_tau)) * np.nan
    np.zeros((n_t, n_ranges)) * np.nan
    pc_pp = {}
    gl_pp = np.zeros((n_t, n_ranges, n_tau)) * np.nan
    gl_pp = {}
    rvalue = np.zeros((n_t, n_tau, n_steps)) * np.nan
    rvalue2 = np.zeros((n_t, n_tau, n_steps)) * np.nan
    range_glue_ext_min = np.zeros((n_t, n_tau, n_steps)) * np.nan
    range_glue_ext_max = np.zeros((n_t, n_tau, n_steps)) * np.nan
    range_glue_ext_thick = np.zeros((n_t, n_tau, n_steps)) * np.nan
    array_gluing_height = np.zeros((n_t, n_tau)) * np.nan

    # Loop over signals
    for j, tau_ in enumerate(tau):
        # Dead time correction photoncounting
        pc_ij = apply_dead_time_correction(pc, tau_)
        pc_pp[j] = pc_ij.copy()

        # Gluing
        gl_pp[j], _, gluing_height = gluing(
            an, pc, range_min=gluing_range[0], range_max=gluing_range[1]
        )

        array_gluing_height[:, j] = gluing_height

    # correlación AN vs PC en la region
    for k in range(n_steps):
        for idx_time, _ in enumerate(times):
            # set_trace()
            for j, tau_ in enumerate(tau):
                # anchura de region
                # Cost Function: para cada perfil uso el mismo rango para todos los taus
                try:
                    min_glue = array_gluing_height[idx_time, j] - exp_glue[k] // 2.0
                    max_glue = array_gluing_height[idx_time, j] + exp_glue[k] // 2.0
                except:
                    raise ValueError("Gluing range out of bounds.")

                idx_glue = np.logical_and(ranges >= min_glue, ranges <= max_glue)
                rvalue[idx_time, j, k] = np.corrcoef(
                    an[idx_time, idx_glue], pc[idx_time, idx_glue]
                )[0, 1]
                if rvalue[idx_time, j, k] >= 0.9:
                    # rangos
                    range_glue_ext_min[idx_time, j, k] = min_glue
                    range_glue_ext_max[idx_time, j, k] = max_glue
                    range_glue_ext_thick[idx_time, j, k] = max_glue - min_glue
                    rvalue2[idx_time, j, k] = np.corrcoef(
                        gl_pp[j][idx_time, idx_glue], pc_pp[j][idx_time, idx_glue]
                    )[0, 1]
                    if rvalue2[idx_time, j, k] >= 0.5:
                        cost_function[idx_time, j, k] = residuals(
                            pc_pp[j][idx_time, idx_glue], gl_pp[j][idx_time, idx_glue]
                        )
                        print(
                            rvalue[idx_time, j, k],
                            rvalue2[idx_time, j, k],
                            cost_function[idx_time, j, k],
                        )

    # Tau that Minimizes Cost Function and associated rvalues for an_pc and gl_pc
    for k in range(n_steps):
        for idx_time, _ in enumerate(times):
            jj = cost_function[idx_time, :, k]
            rr = rvalue[idx_time, :, k]
            rr2 = rvalue2[idx_time, :, k]
            th = range_glue_ext_thick[idx_time, :, k]
            try:
                if not np.isnan(jj).any():
                    idx_Jmin = np.nanargmin(jj)
                    tau_min[idx_time, k] = tau[idx_Jmin]
                    r_min[idx_time, k] = rr[idx_Jmin]
                    r_min2[idx_time, k] = rr2[idx_Jmin]
                    thickness_minJ[idx_time, k] = th[idx_Jmin]
            except Exception as e:
                print(str(e))
                print(
                    "Tau min not estimated for profile %i-th, step %i-th"
                    % (idx_time, k)
                )

            # CRITERIUM TO ESTIMATE DEAD TIME
            # for every thickness, a best tau is estimated above.
            # So, we have a list of best taus. from them, we derive the best
            # of the best, considering the correlation index gl_pc
            # if np.isnan(jj).any() == True:
            #     set_trace()

            try:
                try:
                    idx_best_tau = r_min2[idx_time, :] >= (
                        np.nanmax(r_min2[idx_time, :]) - np.nanstd(r_min2[idx_time, :])
                    )
                except Exception as e:
                    raise ValueError(str(e))

                tau_best_avg[idx_time] = np.nanmean(tau_min[idx_time, idx_best_tau])
                tau_best_std[idx_time] = np.nanstd(tau_min[idx_time, idx_best_tau])
                thickness_best_avg[idx_time] = np.nanmean(
                    thickness_minJ[idx_time, idx_best_tau]
                )
                thickness_best_std[idx_time] = np.nanstd(
                    thickness_minJ[idx_time, idx_best_tau]
                )
            except Exception as e:
                print(str(e))
                print("Best Tau not estimated for profile %i-th" % i)

    # save results in netcdf
    ds = xr.Dataset(
        {
            "J": (["times", "tau", "steps"], cost_function),
            "tau_min": (["times", "steps"], tau_min),
            "tau_best_avg": (["times"], tau_best_avg),
            "tau_best_std": (["times"], tau_best_std),
            "thickness_best_avg": (["times"], thickness_best_avg),
            "thickness_best_std": (["times"], thickness_best_std),
            "range_glue_extended_min": (["times", "tau", "steps"], range_glue_ext_min),
            "range_glue_extended_max": (["times", "tau", "steps"], range_glue_ext_max),
            "range_glue_extended_thick": (
                ["times", "tau", "steps"],
                range_glue_ext_thick,
            ),
            "rvalue_an_pc": (["times", "tau", "steps"], rvalue),
            "rvalue_gl_pc": (["times", "tau", "steps"], rvalue2),
        },
        coords={"times": times, "tau": tau, "steps": exp_glue, "ranges": ranges},
    )
    fn = None
    # fn = os.path.join(config.dead_time_dn, "dead_time_%s_%s_%s_%s_%i.nc"
    #   % (date_str, channel, don_str, gluing_type, expand_bins))
    # ds.to_netcdf(fn)
    # time_elapsed = time.time() - t_start
    # print("%.1f s of time elapsed for %s:%s (%i profiles). channel=%s" % (
    # time_elapsed, date_str, don_str, ds.dims["times"], channel))

    return ds, fn



"""
PLOTTING FUNCTIONS
"""

#Create a function with the following instructions:
#1. The input is a zip filepath with with netcdf files inside.
#2. Check filepath exists. If not return a warning.
#3. Open the zip file and read the netcdf files inside.
#4. Plot the netcdf files.


 

def plot_scc_input(filelist):
    """
    Plot SCC input files.
    Input: 
    filelist: File string format (e.g., /c/*.nc') [str]    
    Output: 
    png figure files. Figures are saved in the same file directory.
    """
    MAX_ALT = 1000 #Points [bins]

    if len(filelist) > 0:
        for file_ in filelist:
            if os.path.isfile(file_):
                lxarray = xr.open_dataset(file_)
                output_directory = os.path.dirname(file_)
                filename = os.path.basename(file_)    

                colorbar = matplotlib.cm.get_cmap('jet', lxarray.time.size)  #Cool        
                colors = colorbar(np.linspace(0, 1, lxarray.time.size))
                try:
                    for idx_channel in np.arange(lxarray.channel_ID.size):                  
                        channel = lxarray.channel_ID.values[idx_channel]
                        fig = plt.figure(figsize=(9,5))
                        try:
                            for idx_time in np.arange(lxarray.time.size):
                                lxarray['Raw_Lidar_Data'][idx_time,idx_channel,0:MAX_ALT].plot(c=colors[idx_time], label='%d' % idx_time)
                        except:
                            logger.warning("Error: in plot_scc_input for %s. Plotting Raw Lidar Data. Try to continue plotting" % file_)
                        try:
                            for idx_time in np.arange(lxarray.time_bck.size):
                                lxarray['Background_Profile'][idx_time,idx_channel,0:MAX_ALT].plot(c=colors[idx_time], label='%d' % idx_time)
                        except:
                            logger.warning("Error: in plot_scc_input for %s. Plotting Background. Try to continue plotting" % file_)
                        plt.title('SCC-channel %d | %s' % (channel, filename))
                        plt.ylabel('Raw_Lidar_Data channel %d [counts/mV]' % channel)
                        plt.xlim(0,MAX_ALT)
                        ymin_value = 0.9*lxarray['Raw_Lidar_Data'][:,idx_channel,0:MAX_ALT].min()
                        ymax_value = 1.1*lxarray['Raw_Lidar_Data'][:,idx_channel,0:MAX_ALT].max()
                        plt.ylim(ymin_value,ymax_value)        
                        plt.savefig(os.path.join(output_directory, '%s_Raw_Lidar_Data_%d.png' % (filename.split('.')[0], channel)), dpi=200)
                        plt.close(fig)
                        logger.info("Plot SCC Raw Data in file %s" % file_)
                except:
                    logger.warning("Error: No plot for %s" % file_)
    else:
        logger.warning('Plot SCC INPUT not Done. Slot File List Empty')


def plot_scc_output(output_dir, scc_code=None):
    """
    Plot SCC optical output products, if available:
    backstter, extinction, angstrom exponent, lidar ratio, particle and volume depolarization

    Parameters:
    ----------
    output_dir: str
        output directory where scc processing products are downloaded
    scc_code: int
        system id

    """

    # config
    font = {'size': 12}
    matplotlib.rc('font', **font)

    # TODO: products id associated to beta, alfa, depo will be read from config file
    #beta_products = [1203, 838, 845, 1199, 760, 669, 839, 863]
    #alfa_products = [850, 1200, 838, 669, 839, 863]
    #depo_products = [760, 1203, 838, 669, 839, 863]
    #inversion_id = {760: 'K', 838: 'K', 845: 'R', 850: 'R', 1199: 'K',
    #                1200: 'R', 1203: 'R', 669: 'K', 839: 'K', 863: 'K'}

    # Plot x,y ranges
    y_lim = (0.68, 10)
    x_lim = {'beta': (-1e-2, 10), 
             'alfa': (1e-2, 500), 
             'angstrom': (-1, 3),
             'depo': (0, 0.40), 
             'lr': (10, 100)}
    # Colors according to Wavelength and Type of retrieval (Klett, K, or Iterative, R)
    colors = {'355': 'tab:blue', '355K': 'dodgerblue', '355I': 'dodgerblue', '355R': 'darkblue', '355U': 'tab:blue',
              '532': 'tab:green', '532K': 'limegreen', '532I': 'limegreen', '532R': 'darkgreen', '532U': 'tab:green',
              '1064': 'tab:red', '1064K': 'red', '1064I': 'red', '1064U': 'tab:red'}

    # necessary variables to be found in scc_optical/pid*nc
    vars_ = ["backscatter", "extinction", "particledepolarization", "volumedepolarization"]

    # scc
    if scc_code is None:
        scc_code = int(
            re.search(r"scc\d{3}", output_dir).group().split("scc")[1])

    logger.info('Plot SCC output. Processing folder: %s' % output_dir)

    # we need all nc files in the directory scc_optical
    nc_files = glob.glob(os.path.join(output_dir, 'scc_optical', '*.nc'))
    if nc_files:
        scc_slot_date_dt = dt.datetime.strptime(re.search(r"/\d{4}/\d{2}/\d{2}/", nc_files[0]).group(), '/%Y/%m/%d/') 
        # Take unique dates in the folder to know the number of different inversions
        def find_dates_in_file(fn, STATION_ID):
            if "pid" in os.path.basename(fn):
                date_pattern = r"\d{2}\d{2}\d{2}\d{2}\d{2}"
                pid_type = 0
            else:  # >= 2021
                date_pattern = r"\d{4}\d{2}\d{2}" + STATION_ID + "\d{2}\d{2}"
                pid_type = 1
            try:
                if pid_type == 0:
                    dates_str = re.search(date_pattern, os.path.basename(fn).split('_')[1].split('.')[0]).group()
                elif pid_type == 1:
                    dates_str = re.search(date_pattern, os.path.basename(fn).split('_')[6]).group()
                    #dates_str = dt.datetime.strptime(dates_str, "%Y%m%d"+STATION_ID+"%H%M").strftime("%Y%m%d%H%M")
                else:
                    dates_str = None
            except Exception as e:
                logger.error("Something went wrong finding date pattern in pid files")
                dates_str = None
            return dates_str
        dates_str = [find_dates_in_file(fn, STATION_ID) for fn in nc_files]
        if dates_str is not None:
            dates_str = np.unique(dates_str).tolist()
        # Loop over dates
        for date_ in dates_str:
            # pid files for the date
            pid_files = glob.glob(os.path.join(output_dir, 'scc_optical', '*%s*.nc' % date_))
            if len(pid_files) > 0:
                # define fig
                fig = plt.figure(figsize=(15, 10))  # , constrained_layout=True)

                """ Store Products in a dictionary """
                profile = {}
                for pid_fn in pid_files:
                    # product id
                    if "pid" in pid_fn:
                        product_id = int(os.path.basename(pid_fn).split('_')[0].replace('pid', ''))
                    else:  # >= scc version 2021
                        product_id = int(os.path.basename(pid_fn).split('_')[3])

                    with xr.open_dataset(pid_fn) as aux_ds:
                        pid_ds = None
                        for i_var in vars_:
                            try:
                                var_ds = aux_ds[i_var].squeeze()  # remove dimensions of length=1
                                # change dimensions for backscatter and extinction
                                if np.logical_or(i_var == "backscatter",
                                                 i_var == "extinction"):
                                    var_ds *= 1e6
                                    var_ds.attrs['units'] = "1/(Mm*sr)"
                                # add to pid dataset
                                if pid_ds is None:
                                    pid_ds = var_ds.to_dataset(name=i_var)
                                else:
                                    pid_ds[i_var] = var_ds
                                # add inversion method as attribute
                                # TODO: particularize for rest of variables: extinction, particledepolarization, volumedepolarization
                                try:
                                    if i_var == "backscatter":
                                        method = aux_ds["backscatter_evaluation_method"].values
                                        if method == 0:
                                            algorithm = aux_ds["raman_backscatter_algorithm"].values
                                        elif method == 1:
                                            algorithm = aux_ds["elastic_backscatter_algorithm"].values
                                        else:
                                            algorithm = 0
                                    else: #if i_var == "extinction":
                                        method = -1 # aux_ds["backscatter_evaluation_method"].values
                                        algorithm = -1 # aux_ds["elastic_backscatter_algorithm"].values
                                    if method == 0:  # Raman
                                        inversion_method = "R"
                                    elif method == 1:  # Elastic Backscatter
                                        if algorithm == 0:
                                            inversion_method = "K"   # Klett-Fernald
                                        elif algorithm == 1:
                                            inversion_method = "I"  # Iterative
                                        else:
                                            inversion_method = ""  # Unknown
                                    else:
                                        inversion_method = ""  # Unknown
                                except Exception as e:
                                    logger.error(str(e))
                                    logger.error("Inversion Method not found")
                                    inversion_method = None
                                pid_ds[i_var].attrs["inversion_method"] = inversion_method
                            except Exception as e:
                                logger.warning("%i does not have %s profile" % (product_id, i_var))
                    # altitude in km
                    pid_ds["altitude"] = pid_ds["altitude"] * 1e-3
                    pid_ds["altitude"].attrs['units'] = "km"

                    # save in profile dictionary
                    profile[product_id] = pid_ds.squeeze()  # remove dimensions of length=1

                """ Plot Backscatter """
                plot_code = 'beta'
                ax = fig.add_subplot(151)
                # loop over pids
                for pid in profile.keys():
                    #if pid in beta_products:  # realmente, ¿hace falta saber si pid está en beta_products?
                    if 'backscatter' in profile[pid].keys():
                        try:
                            wave = int(profile[pid]["wavelength"])
                            inv_met = profile[pid]['backscatter'].attrs["inversion_method"]
                            beta_id = '%d%s' % (wave, inv_met)
                            try:
                                color_beta = colors[beta_id]
                            except:
                                color_beta = colors["%s" % wave]
                            profile[pid]['backscatter'].plot(y='altitude',
                                                             ax=ax, linewidth=2,
                                                             c=color_beta,
                                                             label=beta_id)
                        except Exception as e:
                            logger.warning("Backscatter Not Plotted for %i" % wave)
                ax.xaxis.set_minor_locator(MultipleLocator(1))
                ax.xaxis.grid(b=True, which='minor', linestyle='--')
                ax.set_ylabel(r'Altitude, km asl', fontsize='large')
                ax.set_xlabel(r'$\beta_{a}, Mm^{-1} sr^{-1}$', fontsize='large')
                ax.set_xlim(x_lim[plot_code])
                #ax.set_xscale('log')
                if len(ax.get_lines()) > 0:
                    plt.legend(loc=1, fontsize='medium')

                """ Plot Extinction """
                plot_code = 'alfa'
                ax = fig.add_subplot(152)
                # loop over pids
                for pid in profile.keys():
                    if 'extinction' in profile[pid].keys():
                        try:  # if pid in alfa_products:
                            wave = int(profile[pid]["wavelength"])
                            #inv_met = profile[pid]['extinction'].attrs["inversion_method"]
                            inv_met = ""
                            alfa_id = '%d%s' % (wave, inv_met)
                            try:
                                color_alfa = colors[alfa_id]
                            except:
                                color_alfa = colors["%i" % wave]
                            profile[pid]['extinction'].plot(y='altitude', ax=ax,
                                                            linewidth=2,
                                                            c=color_alfa,
                                                            label=alfa_id)
                        except Exception as e:
                            logger.warning("Extinction Not Plotted for %i" % wave)
                ax.xaxis.set_minor_locator(MultipleLocator(50))
                ax.xaxis.grid(b=True, which='minor', linestyle='--')
                ax.set_xlabel(r'$ \alpha _{a}, Mm^{-1} $', fontsize='large')
                ax.set_xlim(x_lim[plot_code])
                if len(ax.get_lines()) > 0:
                    plt.legend(loc=1, fontsize='medium')

                """ Plot Angstrom """
                ax = fig.add_subplot(153)
                plot_code = 'angstrom'
                angstrom = {}
                if profile != {}:
                    coefficients = ['backscatter', 'extinction']
                    coef_id = {'backscatter': 'beta', 'extinction': 'alpha'}
                    for coef_ in coefficients:
                        for pid_1 in profile.keys():
                            for pid_2 in profile.keys():
                                wave_1 = int(profile[pid_1]['wavelength'])
                                wave_2 = int(profile[pid_2]['wavelength'])
                                #inv_met_1 = profile[pid_1][coef_].attrs["inversion_method"]
                                #inv_met_2 = profile[pid_2][coef_].attrs["inversion_method"]
                                inv_met_1 = ""
                                inv_met_2 = ""
                                if wave_1 < wave_2:
                                    if np.logical_and(
                                            coef_ in profile[pid_1].keys(),
                                            coef_ in profile[pid_2].keys()):
                                        try:
                                            profile[pid_1][coef_][profile[pid_1][coef_] <= 0] = np.nan
                                            profile[pid_2][coef_][profile[pid_2][coef_] <= 0] = np.nan
                                            angstrom_id = r'$\%s$ (%d%s-%d%s)' % (coef_id[coef_], 
                                                    wave_1, inv_met_1, wave_2, inv_met_2)
                                            angstrom[angstrom_id] = (-1) * np.log(
                                                profile[pid_1][coef_] / profile[pid_2][coef_]) / np.log(wave_1 / wave_2)
                                            angstrom[angstrom_id].name = angstrom_id
                                            angstrom[angstrom_id].plot(y='altitude', ax=ax,
                                                                       linewidth=2, label=angstrom_id)
                                        except Exception as e:
                                            logger.warning("Angstrom for %s, %s, %s not plotted. %s" 
                                                    % (coef_, pid_1, pid_2,str(e)))
                    ax.xaxis.set_minor_locator(MultipleLocator(0.5))
                    ax.xaxis.grid(b=True, which='minor', linestyle='--')
                    ax.set_ylim(y_lim)
                    ax.set_xlim(x_lim[plot_code])
                    ax.set_xlabel('Angstrom exponent, #', fontsize='large')
                    if len(ax.get_lines()) > 0:
                        plt.legend(loc=1, fontsize='medium')

                """ Lidar ratio """
                ax = fig.add_subplot(154)
                plot_code = 'lr'
                lr = {}
                if profile != {}:
                    for pid in profile.keys():
                        if np.logical_and('backscatter' in profile[pid].keys(),
                                          'extinction' in profile[pid].keys()):
                            try:
                                wave = int(profile[pid]["wavelength"])
                                #lr_id = '%d%s' % (wave, inversion_id[pid])
                                lr_id = '%d' % (wave,)
                                profile[pid]['extinction'][profile[pid]['extinction'] <= 0] = np.nan
                                profile[pid]['backscatter'][profile[pid]['backscatter'] <= 0] = np.nan
                                lr[lr_id] = profile[pid]['extinction'] / profile[pid]['backscatter']
                                lr[lr_id].name = lr_id
                                lr[lr_id].plot(y='altitude', ax=ax, linewidth=2,
                                               c=colors[lr_id], label=lr_id)
                            except Exception as e:
                                logger.warning("Lidar Ratio for %s not plotted. %s" % (pid, str(e)))
                    ax.xaxis.set_minor_locator(MultipleLocator(10))
                    ax.xaxis.grid(b=True, which='minor', linestyle='--')
                    ax.set_ylim(y_lim)
                    ax.set_xlim(x_lim[plot_code])
                    ax.set_xlabel('Lidar ratio, sr', fontsize='large')
                    if len(ax.get_lines()) > 0:
                        plt.legend(loc=1, fontsize='medium')

                """ Depolarization """
                plot_code = 'depo'
                ax = fig.add_subplot(155)
                for pid in profile.keys():
                    if np.logical_and(
                            "particledepolarization" in profile[pid].keys(),
                            "volumedepolarization" in profile[pid].keys()):# pid in depo_products:
                        wave = int(profile[pid]["wavelength"])
                        #depo_id = '%d%s' % (wave, inversion_id[pid])
                        depo_id = '%d' % (wave,)
                        try:
                            profile[pid]['particledepolarization'].plot(
                                y='altitude', ax=ax, linewidth=2,
                                c=colors[depo_id], label=depo_id)
                            profile[pid]['volumedepolarization'].plot(
                                y='altitude', ax=ax, linewidth=2,
                                linestyle='dashed', c=colors[depo_id], label=depo_id)
                        except Exception as e:
                            logger.warning("Not Plotted Depolarization for %i" % wave)
                ax.xaxis.set_minor_locator(MultipleLocator(0.05))
                ax.xaxis.grid(b=True, which='minor', linestyle='--')
                ax.set_xlabel(r'$\delta_{a}$,#', fontsize='large')
                ax.set_xlim(x_lim[plot_code])
                if len(ax.get_lines()) > 0:
                    plt.legend(loc=1, fontsize='medium')

                # Fig Format and details
                fig_title = '%s' % np.datetime_as_string(profile[pid].time.values, unit='s')
                plt.suptitle(fig_title.replace('T', ' '),verticalalignment='baseline')
                plt.subplots_adjust(bottom=0.25, top=0.95)
                for ax in fig.get_axes():
                    ax.yaxis.set_major_locator(MultipleLocator(1))
                    ax.tick_params(axis='both', labelsize=14)
                    ax.set_ylim(y_lim)
                    ax.set_title('')
                    ax.grid()
                    ax.label_outer()

                # Save Fig
                if profile != {}:
                    fig_dir = os.path.dirname(output_dir)
                    fig_fn = os.path.join(fig_dir, 'scc%d_%s.png' %
                                          (scc_code, fig_title.replace('-', '').replace(':', '').replace('T', '_')))
                    plt.savefig(fig_fn, dpi=200, bbox_inches="tight")
                    plt.close(fig)
                    if os.path.isfile(fig_fn):
                        logger.info('Figure successfully created: %s' % fig_fn)
                        #os.path.basename(fig_fn))
    else:
        logger.warning('ERROR. No NC Files/scc_optical in Folder: %s ' % os.path.basename(output_dir))




import os
import datetime as dt
from pathlib import Path

import numpy as np
import xarray as xr
from gfatpy.lidar.utils.file_manager import filename2info
from gfatpy.lidar.quality_assurance.plot import plot_telecover_channel

from gfatpy.lidar.utils.utils import LIDAR_INFO, signal_to_rcs
from gfatpy.lidar.preprocessing import preprocess

pol_name = {0: "all", 1: "parallel", 2: "perpendicular"}
det_mod_name = {0: "analog", 1: "photoncounting", 2: "gluing"}


def telecover(
    wildcard: str,
    dir: Path | str | None = None,
    output_dir: Path | str | None = None,
    channels: list[str] | None = None,
    normalization_range: tuple[float, float] = (0, 3000),
    crop_range: tuple[float, float] = (0, 3000),
    save_fig: bool = False,
) -> None:

    if dir is None:
        dir = Path.cwd()
    elif isinstance(dir, str):
        dir = Path(dir)

    if not dir.exists() or not dir.is_dir():
        raise ValueError(f"Directory {dir} does not exist or is not a directory.")

    if output_dir is None:
        output_dir = Path.cwd()
    elif isinstance(output_dir, str):
        output_dir = Path(output_dir)

    if not output_dir.exists() or not output_dir.is_dir():
        raise ValueError(
            f"Directory {output_dir} does not exist or is not a directory."
        )

    files = dir.glob(f"{wildcard}.nc")

    # Check files of the same time
    hours = list()  # to list telecover hours (should be the same)
    dates = list()  # to list telecover days (should be the same)
    for file_ in files:
        _, _, _, _, _, date_ = filename2info(file_.name)
        hour_ = file_.name.split("_")[-1].split(".")[0]
        hours.append(hour_)
        dates.append(date_)
    unique_dates = np.unique(dates)  # Number of telecovers of the same type
    unique_hours = np.unique(hours)  # Number of telecovers of the same type
    if len(unique_dates) > 1:
        raise RuntimeError(
            f"Files of different days detected: {dates}. Change the wildcard."
        )
    if len(unique_hours) > 1:
        raise RuntimeError(
            f"Files of different telecovers detected: {hours}. Change the wildcard."
        )

    # Check not repeated sector
    sectors = list()
    for file_ in files:
        sector_ = file_.name.split("_")[1]
        sectors.append(sector_)
    if np.unique(sectors).size < len(sectors):
        raise RuntimeError(f"Repeated sector detected: {sectors}. Change the wildcard.")

    telecover = dict()
    for file_ in files:
        sector_ = file_.name.split("_")[2].split("-")[-1]

        # LIDAR preprocessing
        telecover[sector_] = preprocess(file_)

        # Creating a Dataset [rf_ds] for each channel with all the sectors
        for channel_ in telecover[sector_].channel.values:
            if channel_ in channels and bool(
                telecover[sector_].sel(channel=channel_).active_channel.values.all()
            ):

                # Wavelength, Detection Mode, Polarization
                wavelength = np.floor(
                    telecover[sector_].wavelength.sel(channel=channel_).values
                ).astype("int")
                detection_mode = int(
                    telecover[sector_].detection_mode.sel(channel=channel_).values
                )
                polarization = int(
                    telecover[sector_].polarization.sel(channel=channel_).values
                )

                rf_ds = xr.Dataset()
                for sector_ in enumerate(telecover.keys()):
                    # TELECOVER EARLINET DATA FORMAT
                    # TODO: sacar a una funcion
                    rcs: xr.DataArray = signal_to_rcs(
                        telecover[sector_][f"signal_{channel_}"].mean(
                            dim="time", keep_attrs=True
                        ),
                        telecover[sector_]["range"] ** 2,
                    )
                    rcs.name = sector_

                    rf_ds = xr.merge([rf_ds, rcs], combine_attrs="no_conflicts")

                # Assigning coordinates
                rf_ds = rf_ds.assign_coords({"range": rf_ds.range / 1e3})

                # Attributes
                rf_ds["range"].attrs["units"] = "m"
                rf_ds["range"].attrs["long_name"] = "Height"
                rf_ds = rf_ds.sel(range=slice(*crop_range))
                rf_ds["wavelength"] = wavelength
                rf_ds["wavelength"].attrs["value_str"] = str(wavelength)

                # TODO: sacar la creacion de un dataset especifico para RF a una funcion
                # Polarization
                rf_ds["polarization"] = polarization
                rf_ds["polarization"].attrs["meaning"] = pol_name[polarization]
                rf_ds["polarization"].attrs["id"] = LIDAR_INFO["metadata"][
                    "code_polarization_number2str"
                ][polarization]

                # TODO: sacar esto a una funcion
                # Detection mode
                rf_ds["detection_mode"] = detection_mode
                rf_ds["detection_mode"].attrs["meaning"] = det_mod_name[detection_mode]
                rf_ds["detection_mode"].attrs["id"] = LIDAR_INFO["metadata"][
                    "code_mode_number2str"
                ][detection_mode]

                rf_ds.attrs["sectors"] = list(telecover.keys())
                rf_ds.attrs["lidar_location"] = "Granada"
                rf_ds.attrs["lidar_id"] = "gr"
                rf_ds.attrs["lidar_system"] = telecover[sector_].attrs["system"].lower()
                rf_ds.attrs["datetime_ini"] = dates[0].strftime("%Y%m%dT%H:%M:%S.%f")
                rf_ds.attrs["date_format"] = "%Y%m%dT%H:%M:%S.%f"
                rf_ds.attrs["hour_ini"] = hours[0]
                rf_ds.attrs["hour_end"] = hours[-1]
                rf_ds.attrs["channel_code"] = channel_

                # SAVE RF FILE
                # TODO: sacar a una funcion el escribir este archivo con formato
                cols = list(telecover.keys())
                if detection_mode == 0:  # (if analog)
                    cols.append("D")

                # Convert to pandas
                rf_df = []
                rf_df = rf_ds[cols].to_dataframe()

                # Create file
                lidar_id = rf_ds.attrs["lidar_id"]
                rf_fn = output_dir / f"telecover_{lidar_id}_{channel_}.csv"

                with open(rf_fn, "w") as f:
                    f.write(
                        "station ID = %s (%s)\n"
                        % (
                            rf_ds.attrs["lidar_id"],
                            rf_ds.attrs["lidar_location"],
                        )
                    )
                    f.write("system = %s\n" % rf_ds.attrs["lidar_system"])
                    f.write(
                        "signal = %s, %s, %s, %s\n"
                        % (
                            rf_ds["wavelength"].attrs["value_str"],
                            rf_ds["polarization"].attrs["meaning"],
                            rf_ds["detection_mode"].attrs["meaning"],
                            rf_ds.attrs["dark_subtracted"],
                        )
                    )
                    f.write(
                        "date, time= %s\n"
                        % dt.datetime.strftime(
                            rf_ds.attrs["datetime_ini"], "%d.%m.%Y, %HUTC"
                        )
                    )
                f.close()
                rf_df.index = rf_df.index.map(lambda x: "%.4f" % x)
                rf_df.to_csv(
                    rf_fn,
                    mode="a",
                    header=True,
                    na_rep="NaN",
                    float_format="%.4e",
                )
                if os.path.isfile(rf_fn):
                    # Plot Telecover for current channel
                    plot_telecover_channel(
                        rf_ds,
                        normalization_range=normalization_range,
                        output_dir=output_dir,
                        save_fig=save_fig,
                    )

    return None



import matplotlib

matplotlib.use("agg")
import matplotlib.pyplot as plt

from gfatpy.aeronet import reader, aeronet4lidar, plot, typing, utils

matplotlib.use("agg")

# Plot volumne size Distribution
SMALL_SIZE = 12
MEDIUM_SIZE = 14
BIGGER_SIZE = 18

plt.rc("font", size=MEDIUM_SIZE)  # controls default text sizes
plt.rc("axes", titlesize=MEDIUM_SIZE)  # fontsize of the axes title
plt.rc("axes", labelsize=MEDIUM_SIZE)  # fontsize of the x and y labels
plt.rc("xtick", labelsize=MEDIUM_SIZE)  # fontsize of the tick labels
plt.rc("ytick", labelsize=MEDIUM_SIZE)  # fontsize of the tick labels
plt.rc("legend", fontsize=SMALL_SIZE)  # legend fontsize
plt.rc("figure", titlesize=BIGGER_SIZE)  # fontsize of the figure title


# __all__ = [AERONET_INFO, SMALL_SIZE, MEDIUM_SIZE, BIGGER_SIZE]

__all__ = ["reader", "aeronet4lidar", "plot", "typing", "utils"]

__doc__ = """
This modules allows to manage downloadted AERONET database files from AERONET webpage. 

First, read data and header with `gfatpy.aeronet.reader.reader` and `gfatpy.aeronet.reader.header`, respectively.

"""



from pdb import set_trace
from typing import Any
import numpy as np
import math
import xarray as xr
from pandas.core.arrays import ExtensionArray
from scipy.constants import physical_constants
from gfatpy.atmo.atmo import attenuated_backscatter, transmittance
from gfatpy.atmo.freudenthaler_molecular_properties import molecular_depolarization, f_kbwt, f_kbwc


def retrieve_molecular_extinction_and_backscatter(
    r: np.ndarray[float],
    temperature: np.ndarray[float],
    pressure: np.ndarray[float],
    wavelength: float,
    component: str = "ideal",  
) -> tuple[np.ndarray[np.float64], np.ndarray[np.float64], np.ndarray[np.float64]]:
    r"""Function that gets as input the altitude in [m], the temperature in [K], the pressure in [Pa], and the wavelength in [nm].
    Args:
       - r (np.ndarray[np.float64]): altitude array [m] `gfatpy.atmo.atmo`.
       - temperature (float | np.ndarray[float]): surface temperature or temperature profile [$K$]. If temperature is float, scaled Standard Atmosphere is used.
       - pressure (float | np.ndarray[float]): surface pressure or pressure profile [$Pa$]. If pressure is float, scaled Standard Atmosphere is used.
       - wavelength (float): wavelength [$nm$]
       - component (str, optional): consider the king factor in the molecular lidar ratio (option: `ideal` | `total` | `cabannes`). If idealm $8\pi/3$ otherwise, according to Freudenthaler et al., 2018). Defaults to "ideal".

    Returns:
        tuple[np.ndarray[np.float64], np.ndarray[np.float64], np.float64]: alfasca: Rayleigh extinction [$m^{-1}$], betasca: Rayleigh backscatter [$m^{-1}sr^{-1}$], molecular_lidar_ratio: molecular lidar ratio [$sr$]

    References:
       - Rayleigh Extinction using the King factor [$km^{-1}$]
       - See: Bodhaine et al. 1999. "On Rayleigh Optical Depth Calculations"
       - Freudenthaler et al. 2018. "EARLINET Quality Assurance Tools".
    """                

    # Input units: r [m], T0 [Kelvin], P0 [Pa], lambda [nm]
    # Conversion to the units of this function:
    r = (r / 1e3).astype(np.float64)  # Convert altitude to [km]
    # Altitude is in m, eventhough they say it should be in km

    # pressure = (pressure / 100)
    pressure = (pressure / 100).astype(np.float64)  # Convert pressure to [mb]
    temperature = temperature.astype(np.float64)  # Convert temperature to [K]
    NA = physical_constants["Avogadro constant"][0]  # [mol-1]
    Ns = np.array(NA / 22.4141 * 273.15 / 1013.25 * 1e3 *(pressure / temperature), dtype=np.float64) # [molecules/m3] P[mb], T[K] #Convert to float64 to avoid overflow error in Ns**2
        
    # COMPUTE RAYLEIGH EXTINCTION [km-1]. Eq.(22)-(23)
    lab = wavelength * 1e-3

    # Refractive index. Eq.(21) given at 288.15 K, 1013.25 mb, 360 ppm CO2
    n = 1 + pressure / 1013.25 * 288.15 / temperature * 1e-8 * (
        8060.77 + 2481070 / (132.274 - lab ** (-2)) + 17456.3 / (39.32957 - lab ** (-2))
    )

    # F-factor or King Factor. Eq.(23)(5)(6)
    kingf = (
        78.084 * (1.034 + 3.17e-4 / lab**2)
        + 20.946 * (1.096 + 1.385e-3 / lab**2 + 1.448e-4 / lab**4)
        + 0.934 * 1.00
        + 0.036 * 1.15
    ) / (78.084 + 20.946 + 0.934 + 0.036)

    # Scattering cross section [m2/molecule]
    sigma = (
        8
        * math.pi**3
        * (n**2 - 1) ** 2
        / (3 * (wavelength * 1e-9) ** 4 * Ns**2)
        * kingf
    )
    sigma = 8 * math.pi ** 3 * (n ** 2 - 1) ** 2 / (3 * (wavelength * 1e-9) ** 4 * Ns**2) * kingf

    if component == "total":
        lr_function = f_kbwt
    elif component == "cabannes":
        lr_function = f_kbwc
    elif component == "ideal":
        lr_function = lambda x: 1.0
    else:
        raise ValueError(f"{component} not found.")
    molecular_lidar_ratio = (8.*np.pi/3.) * lr_function(wavelength)
    
    # Extinction [m-1]
    alfasca = sigma * Ns
    # betasca = alfasca / 8.37758041 #From Adolfo's code
    betasca = alfasca / molecular_lidar_ratio

    return alfasca, betasca, molecular_lidar_ratio

def molecular_properties(
    wavelength: float,
    pressure: np.ndarray[Any, np.dtype[np.float64]] | ExtensionArray,
    temperature: np.ndarray[Any, np.dtype[np.float64]] | ExtensionArray,
    heights: np.ndarray[Any, np.dtype[np.float64]] | ExtensionArray,
    times: np.ndarray[Any, np.dtype[np.float64]] | None = None,
    component: str = "ideal",
) -> xr.Dataset:
    """Optical molecular properties of the atmosphere.
    Args:

        - wavelength (float): wavelength of our desired beta molecular attenuated
        - pressure (np.ndarray[Any, np.dtype[np.float64]]): pressure profile
        - temperature (np.ndarray[Any, np.dtype[np.float64]]): temperature profile
        - heights (np.ndarray[Any, np.dtype[np.float64]]): height profile
        - times (np.ndarray[Any, np.dtype[np.float64]] | None, optional): time array. Defaults to None. Note: this is not used in the calculation yet. In development.
        - component (str, optional): _description_. Defaults to "ideal".

    Returns:

        - xr.Dataset: molecular backscatter profile [molecular_beta], molecular extinction profile [molecular_alpha], molecular lidar ratio [molecular_lidar_ratio], molecular depolarization ratio [molecular_depolarization_ratio]
    """

    # molecular backscatter and extinction #
    molecular_extinction, molecular_backscatter, molecular_lidar_ratio = retrieve_molecular_extinction_and_backscatter(heights, temperature, pressure, wavelength, component=component)
    
    #TODO: molecular depolarization ratio defined as ideal? 
    molecular_depolarization_ratio = molecular_depolarization(wavelength, component='cabannes')

    attenuated_molecular_backscatter = (
            molecular_backscatter
            * transmittance(molecular_extinction, heights=heights) ** 2
        )


    if times is None:
        mol_properties = xr.Dataset(
            {
                "molecular_beta": (["range"], molecular_backscatter),
                "molecular_alpha": (["range"], molecular_extinction),
                "molecular_lidar_ratio": ([], molecular_lidar_ratio),
                "molecular_depolarization": ([], molecular_depolarization_ratio),
                "atten_molecular_beta": (["range"], attenuated_molecular_backscatter),
            },
            coords={"range": heights},
        )
    else:
        mol_properties = xr.Dataset(
            {
                "molecular_beta": (["time", "range"], molecular_backscatter),
                "molecular_alpha": (["time", "range"], molecular_extinction),          
                "atten_molecular_beta": (["time", "range"], attenuated_molecular_backscatter),      
                "molecular_lidar_ratio": (["time"], molecular_lidar_ratio),
                "molecular_depolarization": (["time"], molecular_depolarization_ratio)

            },
            coords={
                "time": times,
                "range": heights,
            },  # FIXME: time is not used in the calculation yet
        )
    return mol_properties


from pdb import set_trace
import re
import yaml
import zipfile
import fnmatch
import tempfile
import numpy as np
from typing import Any
from pathlib import Path
from datetime import datetime

from gfatpy.utils.utils import datetime_np2dt


def read_yaml(path: Path | str) -> Any:
    """
    Reads a yaml file and returns the data as a dictionary.
    """
    with open(path, "r") as stream:
        return yaml.safe_load(stream)


def read_yaml_from_info(lidar_name: str, target_date: datetime) -> Any:

    file = find_nearest_filepath(
        dir=Path(__file__).parent.parent.absolute() / "lidar" / "info",
        wildcard_filename=f"info_{lidar_name}*.yml",
        date_location_in_filename=2,
        date=target_date,
        and_previous=True,
    )
    return read_yaml(file)


def find_nearest_filepath(
    dir: Path,
    wildcard_filename: str,
    date_location_in_filename: int,
    date: datetime,
    and_previous: bool = False,
) -> Path:
    """Finds the nearest file path in the specified directory based on a wildcard filename pattern and a target date.

    Args:
        dir (Path): The directory to search for files.
        wildcard_filename (str): The wildcard filename pattern to match files.
        date_location_in_filename (int): The index of the date in the filename split by underscores.
        date (datetime): The target date to find the nearest file.
        and_previous (bool, optional): If True, then find the nearest file before the target date. Defaults to False.

    Returns:
        Path: The path to the nearest file.

    Raises:
        ValueError: If no file is found in the directory.

    """
    # Get all candidate file paths
    candidates = np.array([*dir.rglob(wildcard_filename)])
    # Raise an error if no candidate files are found
    if len(candidates) == 0:
        raise ValueError(
            f"No files found in {dir} with the pattern {wildcard_filename}."
        )
    # Extract the dates from candidate file names
    dates_string = [
        p.name.split(".")[0].split("_")[date_location_in_filename] for p in candidates
    ]

    # Convert dates_string to numpy datetime64 objects
    dates = np.array(
        [datetime.strptime(date_str, "%Y%m%d") for date_str in dates_string]
    )

    # if and_previous is True, then find the nearest date before date_target
    if and_previous:
        candidates = candidates[dates < datetime_np2dt(date)]
        dates = dates[dates < datetime_np2dt(date)]

    # Calculate absolute differences between each date and date_target
    date_diffs = np.abs(dates - datetime_np2dt(date))

    # Find the index of the minimum absolute difference
    try:
        nearest_date_index = np.argmin(date_diffs)
    except ValueError:
        raise ValueError("No file found.")

    # Get the path of the nearest file
    path = candidates[nearest_date_index]

    # Raise an error if the path does not exist
    if not path.exists():
        raise ValueError("No file found.")

    return path


def unzip_file(
    file: Path,
    pattern_or_list: str | list[str] = "*.*",
    destination: Path | None = None,
) -> Path:
    """Unzips a zip file to a temporary directory.

    Args:
        file (Path): The path to the zip file.
        pattern (str, optional): The pattern to match files to extract. Defaults to '*.*'.
        destination (Path | None, optional): The destination directory to extract the files. Defaults to None creating a temporary directory.

    Raises:
        ValueError: file is not a zip file.
        ValueError: file is not a valid zip file.
        ValueError: An error occurred.

    Returns:
        Path: The path to the unzipped directory.
    """
    if not file.suffix.endswith("zip"):
        raise ValueError(f"The file {file} is not a zip file.")

    # Create a temporary directory
    if destination is None:
        unzip_file = Path(
            tempfile.mkdtemp(prefix="tmp_unzipped_", dir=Path.cwd().as_posix())
        )
    else:
        unzip_file = Path(tempfile.mkdtemp(prefix="tmp_unzipped_", dir=destination))

    with zipfile.ZipFile(file, "r") as zip_ref:
        if isinstance(pattern_or_list, list):
            files_to_extract = [
                file
                for file in zip_ref.namelist()
                if any(pattern in file for pattern in set(pattern_or_list))
            ]
        elif isinstance(pattern_or_list, str):
            if pattern_or_list == "*.*":
                files_to_extract = zip_ref.namelist()
            else:
                pattern = re.compile(pattern_or_list)
                files_to_extract = [
                    file_info.filename
                    for file_info in zip_ref.infolist()
                    if bool(re.search(pattern, file_info.filename))
                ]
        if len(files_to_extract) == 0:
            raise FileNotFoundError(
                f"Zip file {file}: No files found with pattern/s ({pattern_or_list})."
            )

        for file_ in files_to_extract:
            zip_ref.extract(file_, path=unzip_file)

    return unzip_file
